import logging
import numpy as np
from pathlib import Path
from typing import List, Optional, Union

import autoarray as aa
from autoarray.dataset.abstract.dataset import AbstractDataset

logger = logging.getLogger(__name__)


class DatasetQuantity(AbstractDataset):
    def __init__(
        self,
        data: Union[aa.Array2D, aa.VectorYX2D],
        noise_map: Union[aa.Array2D, aa.VectorYX2D],
        over_sampling: Optional[aa.OverSamplingDataset] = aa.OverSamplingDataset(),
    ):
        """
        A quantity dataset, which represents a derived quantity of a light profile, mass profile, galaxy or galaxies
        that can be fitted with a model via a non-linear.

        For example, the `DatasetQuantity` could represent the `convergence` of a mass profile, and this dataset could
        then be fit with a galaxies via the `AnalysisQuantity` class. The benefit of doing this is that the components
        of different models can be fitted to one another and matched for example, the power-law model whose convergence
        best matches the convergence of a dark matter profile could be inferred).

        The following quantities of the data are available and used for the following tasks:

        - `data`: The quantity data, which shows the signal that is analysed and fitted with a model data of a model
        object.

        - `noise_map`: The RMS standard deviation error in every pixel, which is used to compute the chi-squared value
        and likelihood of a fit.

        Datasets also contains following properties:

        - `grid`: A grids of (y,x) coordinates which align with the image pixels, whereby each coordinate corresponds to
        the centre of an image pixel. This may be used in fits to calculate the model image of the imaging data.

        - `grids.pixelization`: A grid of (y,x) coordinates which align with the pixels of a pixelization. This grid
        is specifically used for pixelizations computed via the `invserion` module, which often use different
        oversampling and sub-size values to the grid above.

        The `over_sampling` and `over_sampling_pixelization` define how over sampling is performed for these grids.

        This is used in the project PyAutoGalaxy to load imaging data of a galaxy and fit it with galaxy light profiles.
        It is used in PyAutoLens to load imaging data of a strong lens and fit it with a lens model.

        Parameters
        ----------
        data
            The data of the quantity (e.g. 2D convergence, 2D potential, 2D deflections) that is fitted.
        noise_map
            The 2D noise map of the quantity's data.
        noise_map
            An array describing the RMS standard deviation error in each pixel used for computing quantities like the
            chi-squared in a fit, which is often chosen in an arbitrary way for a quantity dataset given the quantities
            are not observed using real astronomical instruments.
        over_sampling
            The over sampling schemes which divide the grids into sub grids of smaller pixels within their host image
            pixels when using the grid to evaluate a function (e.g. images) to better approximate the 2D line integral
            This class controls over sampling for all the different grids (e.g. `grid`, `grids.pixelization).
        """
        if data.shape != noise_map.shape:
            if data.shape[0:-1] == noise_map.shape[0:]:
                noise_map = aa.VectorYX2D.no_mask(
                    values=np.stack((noise_map, noise_map), axis=-1),
                    pixel_scales=data.pixel_scales,
                    shape_native=data.shape_native,
                    origin=data.origin,
                )

        self.unmasked = None

        super().__init__(
            data=data,
            noise_map=noise_map,
            over_sampling=over_sampling,
        )

    @classmethod
    def via_signal_to_noise_map(
        cls,
        data: Union[aa.Array2D, aa.VectorYX2D],
        signal_to_noise_map: Union[aa.Array2D],
        over_sampling: Optional[aa.OverSamplingIterate] = None,
    ):
        """
        Represents a derived quantity of a light profile, mass profile, galaxy or galaxies as a dataset that can be
        fitted with a model via a non-linear (see `DatasetQuantity.__init__`).

        This classmethod takes as input a signal-to-noise map, as opposed to the noise map used in the `__init__`
        constructor. The noise-map is then derived from this signal-to-noise map, such that this is the signal to
        noise of the `DatasetQuantity` that is returned.

        Parameters
        ----------
        data
            The data of the quantity (e.g. 2D convergence, 2D potential, 2D deflections) that is fitted.
        signal_to_noise_map
            The 2D signal to noise map of the quantity's data.
        """
        try:
            noise_map = data / signal_to_noise_map
        except ValueError:
            noise_map = aa.VectorYX2D.zeros(
                shape_native=data.shape_native, pixel_scales=data.pixel_scales
            )
            noise_map = noise_map.apply_mask(mask=data.mask)

            signal_to_noise_map[signal_to_noise_map < 1e-8] = 1e-8

            noise_map[:, 0] = np.abs(data.slim[:, 0]) / signal_to_noise_map
            noise_map[:, 1] = np.abs(data.slim[:, 1]) / signal_to_noise_map

        return DatasetQuantity(
            data=data,
            noise_map=noise_map,
            over_sampling=over_sampling,
        )

    @property
    def y(self) -> "DatasetQuantity":
        """
        If the `DatasetQuantity` contains a `VectorYX2D` as its data, this property returns a new `DatasetQuantity`
        with just the y-values of the vectors as the data, alongside the noise-map.

        This is primarily used for visualizing a fit to the `DatasetQuantity` containing vectors, as it allows one to
        reuse tools which visualize `Array2D` objects.
        """
        if isinstance(self.data, aa.VectorYX2D):
            return DatasetQuantity(
                data=self.data.y,
                noise_map=self.noise_map.y,
                over_sampling=self.over_sampling,
            )

    @property
    def x(self) -> "DatasetQuantity":
        """
        If the `DatasetQuantity` contains a `VectorYX2D` as its data, this property returns a new `DatasetQuantity`
        with just the x-values of the vectors as the data, alongside the noise-map.

        This is primarily used for visualizing a fit to the `DatasetQuantity` containing vectors, as it allows one to
        reuse tools which visualize `Array2D` objects.
        """
        if isinstance(self.data, aa.VectorYX2D):
            return DatasetQuantity(
                data=self.data.x,
                noise_map=self.noise_map.x,
                over_sampling=self.over_sampling,
            )

    def apply_mask(self, mask: aa.Mask2D) -> "DatasetQuantity":
        """
        Apply a mask to the quantity dataset, whereby the mask is applied to the data and noise-map one-by-one.

        The original unmasked qunatity dataset is stored as the `self.unmasked` attribute. This is used to ensure that
        if the `apply_mask` function is called multiple times, every mask is always applied to the original unmasked
        imaging dataset.

        Parameters
        ----------
        mask
            The 2D mask that is applied to the image.
        """
        if self.data.mask.is_all_false:
            unmasked_dataset = self
        else:
            unmasked_dataset = self.unmasked

        data = self.data.apply_mask(mask=mask)
        noise_map = self.noise_map.apply_mask(mask=mask)

        dataset = DatasetQuantity(
            data=data,
            noise_map=noise_map,
            over_sampling=self.over_sampling,
        )

        dataset.unmasked = unmasked_dataset

        logger.info(
            f"IMAGING - Data masked, contains a total of {mask.pixels_in_mask} image-pixels"
        )

        return dataset

    @property
    def shape_native(self):
        return self.data.shape_native

    @property
    def pixel_scales(self):
        return self.data.pixel_scales

    def output_to_fits(
        self,
        data_path: Union[Path, str],
        noise_map_path: Optional[Union[Path, str]] = None,
        overwrite: bool = False,
    ):
        """
        Output a quantity dataset to multiple .fits file.

        For each attribute of the imaging data (e.g. `data`, `noise_map`) the path to
        the .fits can be specified, with `hdu=0` assumed automatically.

        If the `data` has been masked, the masked data is output to .fits files. A mask can be separately output to
        a file `mask.fits` via the `Mask` objects `output_to_fits` method.

        Parameters
        ----------
        data_path
            The path to the data .fits file where the image data is output (e.g. '/path/to/data.fits').
        noise_map_path
            The path to the noise_map .fits where the noise_map is output (e.g. '/path/to/noise_map.fits').
        overwrite
            If `True`, the .fits files are overwritten if they already exist, if `False` they are not and an
            exception is raised.
        """
        self.data.output_to_fits(file_path=data_path, overwrite=overwrite)

        if self.noise_map is not None and noise_map_path is not None:
            self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)



from autogalaxy.aggregator.galaxies import GalaxiesAgg
from autogalaxy.aggregator.imaging.imaging import ImagingAgg
from autogalaxy.aggregator.imaging.fit_imaging import FitImagingAgg
from autogalaxy.aggregator.interferometer.interferometer import InterferometerAgg
from autogalaxy.aggregator.interferometer.fit_interferometer import FitInterferometerAgg
from autogalaxy.aggregator.ellipse.ellipses import EllipsesAgg
from autogalaxy.aggregator.ellipse.multipoles import MultipolesAgg
from autogalaxy.aggregator.ellipse.fit_ellipse import FitEllipseAgg



from typing import Tuple

from autogalaxy.profiles.mass.total.power_law_core import PowerLawCore
from autogalaxy.profiles.mass.total.power_law_core import PowerLawCoreSph


class IsothermalCore(PowerLawCore):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        einstein_radius: float = 1.0,
        core_radius: float = 0.01,
    ):
        """
        Represents a cored elliptical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope: float = 2.0

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        einstein_radius
            The arc-second Einstein radius.
        core_radius
            The arc-second radius of the inner core.
        """
        super().__init__(
            centre=centre,
            ell_comps=ell_comps,
            einstein_radius=einstein_radius,
            slope=2.0,
            core_radius=core_radius,
        )


class IsothermalCoreSph(PowerLawCoreSph):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        einstein_radius: float = 1.0,
        core_radius: float = 0.01,
    ):
        """
        Represents a cored spherical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope: float = 2.0

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        einstein_radius
            The arc-second Einstein radius.
        core_radius
            The arc-second radius of the inner core.
        """
        super().__init__(
            centre=centre,
            einstein_radius=einstein_radius,
            slope=2.0,
            core_radius=core_radius,
        )



from autogalaxy.analysis.result import Result
from autogalaxy.quantity.fit_quantity import FitQuantity


class ResultQuantity(Result):
    @property
    def max_log_likelihood_fit(self) -> FitQuantity:
        """
        An instance of a `FitQuantity` corresponding to the maximum log likelihood model inferred by the non-linear
        search.
        """

        return self.analysis.fit_quantity_for_instance(instance=self.instance)



from autoconf import cached_property

from autogalaxy.analysis.adapt_images.adapt_images import AdaptImages


class AdaptImageMaker:
    def __init__(self, result, use_model_images: bool = False):
        """
        Class used for making adapt images on-the-fly for efficient computations.

        Adapt images can be slow to compute, especially for large datasets. When an analysis pipeline is resumed
        from already computed results, the adapt images in early fits may not be used. By default, a pipeline would
        still recalculate the adapt images for every fit, slowing down the time it takes to resume a pipeline.

        By using the `AdaptImageMaker`, the adapt images are computed on-the-fly during the fit, when and only when
        the `adapt_images` attribute of the `Analysis` class is called. This means that if the adapt images are not
        used when resuming a pipeline, they are not recalculated, speeding up the time it takes to resume a pipeline.

        Parameters
        ----------
        result
            The result from a previous stage of the pipeline which contains the information necessary to make the
            adapt images.
        use_model_images
            Whether to use the model images from the result to make the adapt images. If `False`, the galaxies
            subtracted images of each dataset are used.
        """

        self.result = result
        self.use_model_images = use_model_images

    @cached_property
    def adapt_images(self):
        """
        Returns the adapt images from the result.

        The adapt images are therefore only computed when this attribute is called, which for pipelines resuming
        from already computed results means this function is often omitted, saving computational time.

        This function is cached, therefore once the adapt images are computed they are stored in memory and not
        recomputed when this attribute is called again, saving computational time.

        Returns
        -------
        The adapt images from the result.
        """
        return self.result.adapt_images_from(use_model_images=self.use_model_images)



import logging
import numpy as np
from os import path
from typing import Dict, Optional, List

import autofit as af
import autoarray as aa

logger = logging.getLogger(__name__)

logger.setLevel(level="INFO")


class Preloads(aa.Preloads):
    def __init__(
        self,
        w_tilde: Optional[aa.WTildeImaging] = None,
        use_w_tilde: Optional[bool] = None,
        blurred_image: Optional[aa.Array2D] = None,
        image_plane_mesh_grid_pg_list: Optional[List[List[aa.Grid2D]]] = None,
        relocated_grid: Optional[aa.Grid2D] = None,
        mapper_list: Optional[aa.AbstractMapper] = None,
        mapper_galaxy_dict: Optional[Dict[aa.AbstractMapper, "Galaxy"]] = None,
        linear_func_operated_mapping_matrix_dict=None,
        data_linear_func_matrix_dict=None,
        mapper_operated_mapping_matrix_dict=None,
        operated_mapping_matrix: Optional[np.ndarray] = None,
        curvature_matrix: Optional[np.ndarray] = None,
        regularization_matrix: Optional[np.ndarray] = None,
        log_det_regularization_matrix_term: Optional[float] = None,
        traced_mesh_grids_list_of_planes=None,
        image_plane_mesh_grid_list=None,
        failed=False,
    ):
        """
        Class which offers a concise API for settings up the preloads, which before a model-fit are set up via
        a comparison of two fits using two different models. If a quantity in these two fits is identical, it does
        not change thoughout the model-fit and can therefore be preloaded to avoid computation, speeding up
        the analysis.

        For example, the image-plane source-plane pixelization grid (which may be computationally expensive to compute
        via a KMeans algorithm) does not change for the majority of model-fits, because the associated model parameters
        are fixed. Preloading avoids rerruning the KMeans algorithm for every model fitted, by preloading it in memory
        and using this preload in every fit.

        Parameters
        ----------
        blurred_image
            The preloaded array of values containing the blurred image of a model fit (e.g. that light profile of
            every galaxy in the model). This can be preloaded when no light profiles in the model vary.
        w_tilde
            A class containing values that enable an inversion's linear algebra to use the w-tilde formalism. This can
            be preloaded when no component of the model changes the noise map (e.g. galaxies are fixed).
        use_w_tilde
            Whether to use the w tilde formalism, which superseeds the value in `SettingsInversions` such that w tilde
            will be disabled for model-fits it is not applicable (e.g. because the noise-map changes).
        traced_grids_of_planes_for_inversion
            The two dimensional grids corresponding to the traced grids in a lens fit. This can be preloaded when no
             mass profiles in the model vary.
        image_plane_mesh_grid_pg_list
            The two dimensional grids corresponding to the sparse image plane grids in a lens fit, that is ray-traced to
            the source plane to form the source pixelization. This can be preloaded when no pixelizations in the model
            vary.
        relocated_grid
            The two dimensional grids corresponding to the grid that has had its border pixels relocated for a
            pixelization in a lens fit. This can be preloaded when no mass profiles in the model vary.
        mapper_list
            The mapper of a fit, which preloading avoids recalculation of the mapping matrix and image to source
            pixel mappings. This can be preloaded when no pixelizations in the model vary.
        operated_mapping_matrix
            A matrix containing the mappings between PSF blurred image pixels and source pixels used in the linear
            algebra of an inversion. This can be preloaded when no mass profiles and pixelizations in the model vary.

        Returns
        -------
        The preloads object used to skip certain calculations in the log likelihood function.
        """
        super().__init__(
            w_tilde=w_tilde,
            use_w_tilde=use_w_tilde,
            relocated_grid=relocated_grid,
            image_plane_mesh_grid_pg_list=image_plane_mesh_grid_pg_list,
            mapper_list=mapper_list,
            linear_func_operated_mapping_matrix_dict=linear_func_operated_mapping_matrix_dict,
            data_linear_func_matrix_dict=data_linear_func_matrix_dict,
            mapper_operated_mapping_matrix_dict=mapper_operated_mapping_matrix_dict,
            operated_mapping_matrix=operated_mapping_matrix,
            curvature_matrix=curvature_matrix,
            regularization_matrix=regularization_matrix,
            log_det_regularization_matrix_term=log_det_regularization_matrix_term,
            traced_mesh_grids_list_of_planes=traced_mesh_grids_list_of_planes,
            image_plane_mesh_grid_list=image_plane_mesh_grid_list,
        )

        self.mapper_galaxy_dict = mapper_galaxy_dict
        self.blurred_image = blurred_image
        self.failed = failed

    @classmethod
    def setup_all_via_fits(cls, fit_0, fit_1) -> "Preloads":
        """
        Setup the Preloads from two fits which use two different lens model of a model-fit.

        Parameters
        ----------
        fit_0
            The first fit corresponding to a model with a specific set of unit-values.
        fit_1
            The second fit corresponding to a model with a different set of unit-values.

        Returns
        -------
        Preloads
            Preloads which are set up based on the fit's passed in specific to a lens model.

        """

        preloads = cls()

        if isinstance(fit_0, aa.FitImaging):
            preloads.set_w_tilde_imaging(fit_0=fit_0, fit_1=fit_1)
            preloads.set_blurred_image(fit_0=fit_0, fit_1=fit_1)

        preloads.set_mapper_list(fit_0=fit_0, fit_1=fit_1)

        if preloads.mapper_list is not None:
            preloads.mapper_galaxy_dict = fit_0.galaxies_to_inversion.mapper_galaxy_dict

        preloads.set_operated_mapping_matrix_with_preloads(fit_0=fit_0, fit_1=fit_1)
        preloads.set_linear_func_inversion_dicts(fit_0=fit_0, fit_1=fit_1)
        preloads.set_curvature_matrix(fit_0=fit_0, fit_1=fit_1)
        preloads.set_regularization_matrix_and_term(fit_0=fit_0, fit_1=fit_1)

        return preloads

    def set_blurred_image(self, fit_0, fit_1):
        """
        If the `LightProfile`'s in a model are all fixed parameters their corresponding image and therefore PSF blurred
        image do not change during the model fit and can therefore be preloaded.

        This function compares the blurred image of two fit's corresponding to two model instances, and preloads
        the blurred image if the blurred image of both fits are the same.

        The preload is typically used though out search chaining pipelines, as it is common to fix the lens light for
        the majority of model-fits.

        Parameters
        ----------
        fit_0
            The first fit corresponding to a model with a specific set of unit-values.
        fit_1
            The second fit corresponding to a model with a different set of unit-values.
        """
        self.blurred_image = None

        precision = 1e-8

        if (np.max(abs(fit_0.blurred_image - fit_1.blurred_image)) < precision) and (
            np.sum(fit_0.blurred_image) > precision
        ):
            self.blurred_image = fit_0.blurred_image

            logger.info(
                "PRELOADS - Blurred image (e.g. the image of all light profiles) is preloaded for this model-fit."
            )

    def output_info_to_summary(self, file_path):
        file_preloads = path.join(file_path, "preloads.summary")

        if self.failed:
            logger.info(
                "PRELOADS - Preloading failed as models gave too many exceptions, preloading therefore "
                "not used."
            )

            af.formatter.output_list_of_strings_to_file(
                file=file_preloads, list_of_strings=["FAILED"]
            )

        else:
            af.formatter.output_list_of_strings_to_file(
                file=file_preloads, list_of_strings=self.info
            )

    @property
    def info(self) -> List[str]:
        """
        The information on what has or has not been preloaded, which is written to the file `preloads.summary`.

        Returns
        -------
            A list of strings containing True and False values as to whether a quantity has been preloaded.
        """
        line = [f"W Tilde = {self.w_tilde is not None}\n"]
        line += [f"Use W Tilde = {self.use_w_tilde}\n\n"]
        line += [f"Blurred Image = {np.count_nonzero(self.blurred_image) != 0}\n"]
        line += [f"Mapper = {self.mapper_list is not None}\n"]
        line += [
            f"Blurred Mapping Matrix = {self.operated_mapping_matrix is not None}\n"
        ]
        line += [
            f"Inversion Linear Func (Linear Light Profile) Dicts = {self.linear_func_operated_mapping_matrix_dict is not None}\n"
        ]
        line += [f"Curvature Matrix = {self.curvature_matrix is not None}\n"]
        line += [
            f"Curvature Matrix Mapper Diag = {self.curvature_matrix_mapper_diag is not None}\n"
        ]
        line += [f"Regularization Matrix = {self.regularization_matrix is not None}\n"]
        line += [
            f"Log Det Regularization Matrix Term = {self.log_det_regularization_matrix_term is not None}\n"
        ]

        return line



import autoarray as aa

from typing import List

from autogalaxy.analysis.adapt_images.adapt_images import AdaptImages
from autogalaxy.analysis.result import ResultDataset
from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.interferometer.fit_interferometer import FitInterferometer


class ResultInterferometer(ResultDataset):
    @property
    def max_log_likelihood_fit(self) -> FitInterferometer:
        """
        An instance of a `FitInterferometer` corresponding to the maximum log likelihood model inferred by the
        non-linear search.
        """
        return self.analysis.fit_from(instance=self.instance)

    @property
    def max_log_likelihood_galaxies(self) -> List[Galaxy]:
        """
        An instance of the galaxies corresponding to the maximum log likelihood model inferred by the non-linear search.

        The galaxies are computed from the `max_log_likelihood_fit`, as this ensures that all linear light profiles
        are converted to normal light profiles with their `intensity` values updated.
        """
        return (
            self.max_log_likelihood_fit.model_obj_linear_light_profiles_to_light_profiles
        )

    @property
    def real_space_mask(self) -> aa.Mask2D:
        """
        The real space mask used by this model-fit.
        """
        return self.max_log_likelihood_fit.dataset.real_space_mask

    def adapt_images_from(self, use_model_images: bool = False) -> AdaptImages:
        """
        Returns the adapt-images which are used to make a pixelization's mesh and regularization adapt to the
        reconstructed galaxy's morphology.

        This can use either:

        - The model image of each galaxy in the best-fit model.
        - The subtracted image of each galaxy in the best-fit model, where the subtracted image is the dataset
          minus the model images of all other galaxies.

        In **PyAutoLens** these adapt images have had lensing calculations performed on them and therefore for source
        galaxies are their lensed model images in the image-plane.

        Parameters
        ----------
        use_model_images
            If True, the model images of the galaxies are used to create the adapt images. If False, the subtracted
            images of the galaxies are used.
        """

        return AdaptImages.from_result(
            result=self,
            use_model_images=True,
        )



import numpy as np
from scipy.special import factorial, genlaguerre
from typing import Optional, Tuple

import autoarray as aa

from autogalaxy.profiles.light import standard as lp

from autogalaxy.profiles.light.decorators import (
    check_operated_only,
)
from autogalaxy.profiles.light.linear.abstract import LightProfileLinear


class ShapeletExponential(lp.ShapeletExponential, LightProfileLinear):
    def __init__(
        self,
        n: int,
        m: int,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        beta: float = 1.0,
    ):
        """
        Shapelets where the basis function is defined according to an Exponential using a polar (r,theta) grid of
        coordinates.

        Shapelets are defined according to:

          https://arxiv.org/abs/astro-ph/0105178

        Shapelets are described in the context of strong lens modeling in:

          https://ui.adsabs.harvard.edu/abs/2016MNRAS.457.3066T/abstract

        Parameters
        ----------
        n
            The n order of the shapelets basis function.
        m
            The m order of the shapelets basis function in the x-direction.
        centre
            The (y,x) arc-second coordinates of the profile (shapelet) centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        beta
            The characteristic length scale of the shapelet basis function, defined in arc-seconds.
        """

        super().__init__(n=n, m=m, centre=centre, ell_comps=ell_comps, beta=beta)

    @aa.over_sample
    @aa.grid_dec.to_array
    @check_operated_only
    @aa.grid_dec.transform
    @aa.grid_dec.relocate_to_radial_minimum
    def image_2d_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None, **kwargs
    ) -> np.ndarray:
        """
        Returns the Exponential Shapelet light profile's 2D image from a 2D grid of Exponential (y,x) coordinates.

        If the coordinates have not been transformed to the profile's geometry (e.g. translated to the
        profile `centre`), this is performed automatically.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        image
            The image of the Exponential Shapelet evaluated at every (y,x) coordinate on the transformed grid.
        """

        radial = (grid[:, 0] ** 2 + grid[:, 1] ** 2) / self.beta
        theta = np.arctan(grid[:, 1] / grid[:, 0])

        prefactor = (
            1.0
            / np.sqrt(2 * np.pi)
            / self.beta
            * (self.n + 0.5) ** (-1 - np.abs(self.m))
            * (-1) ** (self.n + self.m)
            * np.sqrt(
                factorial(self.n - np.abs(self.m)) / 2 * self.n
                + 1 / factorial(self.n + np.abs(self.m))
            )
        )

        laguerre = genlaguerre(n=self.n - np.abs(self.m), alpha=2 * np.abs(self.m))
        shapelet = laguerre(radial / (self.n + 0.5))

        return np.abs(
            prefactor
            * np.exp(-radial / (2 * self.n + 1))
            * radial ** (np.abs(self.m))
            * shapelet
            * np.cos(self.m * theta)
            + -1.0j * np.sin(self.m * theta)
        )


class ShapeletExponentialSph(ShapeletExponential):
    def __init__(
        self,
        n: int,
        m: int,
        centre: Tuple[float, float] = (0.0, 0.0),
        beta: float = 1.0,
    ):
        """
        Shapelets where the basis function is defined according to a Exponential (r,theta) grid of coordinates.

        Shapelets are defined according to:

          https://arxiv.org/abs/astro-ph/0105178

        Shapelets are described in the context of strong lens modeling in:

          https://ui.adsabs.harvard.edu/abs/2016MNRAS.457.3066T/abstract

        Parameters
        ----------
        n_y
            The order of the shapelets basis function in the y-direction.
        n_x
            The order of the shapelets basis function in the x-direction.
        centre
            The (y,x) arc-second coordinates of the profile (shapelet) centre.
        beta
            The characteristic length scale of the shapelet basis function, defined in arc-seconds.
        """

        super().__init__(n=n, m=m, centre=centre, ell_comps=(0.0, 0.0), beta=beta)



from functools import partial
from typing import List, Optional

import autofit as af
import autoarray as aa


def _interferometer_from(
    fit: af.Fit,
    real_space_mask: Optional[aa.Mask2D] = None,
) -> List[aa.Interferometer]:
    """
    Returns a list of `Interferometer` objects from a `PyAutoFit` sqlite database `Fit` object.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The interferometer visibilities data as a .fits file (`dataset/data.fits`).
    - The visibilities noise-map as a .fits file (`dataset/noise_map.fits`).
    - The uv wavelengths as a .fits file (`dataset/uv_wavelengths.fits`).
    - The real space mask defining the grid of the interferometer for the FFT (`dataset/real_space_mask.fits`).
    - The settings of the `Interferometer` data structure used in the fit (`dataset/settings.json`).

    Each individual attribute can be loaded from the database via the `fit.value()` method.

    This method combines all of these attributes and returns a `Interferometer` object, including having its
    settings updated to the values used by the model-fit.

    If multiple `Interferometer` objects were fitted simultaneously via analysis summing, the `fit.child_values()`
    method is instead used to load lists of the data, noise-map, PSF and mask and combine them into a list of
    `Interferometer` objects.

    Parameters
    ----------
    fit
        A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
    """

    fit_list = [fit] if not fit.children else fit.children

    dataset_list = []

    for fit in fit_list:
        data = aa.Visibilities(
            visibilities=fit.value(name="dataset.data").data.astype("float")
        )
        noise_map = aa.VisibilitiesNoiseMap(
            fit.value(name="dataset.noise_map").data.astype("float")
        )
        uv_wavelengths = fit.value(name="dataset.uv_wavelengths").data

        real_space_mask = (
            real_space_mask
            if real_space_mask is not None
            else aa.Mask2D.from_primary_hdu(
                primary_hdu=fit.value(name="dataset.real_space_mask")
            )
        )

        over_sampling = fit.value(name="dataset.over_sampling")
        transformer_class = fit.value(name="dataset.transformer_class")

        dataset = aa.Interferometer(
            data=data,
            noise_map=noise_map,
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            over_sampling=over_sampling,
            transformer_class=transformer_class,
        )

        dataset_list.append(dataset)

    return dataset_list


class InterferometerAgg:
    def __init__(self, aggregator: af.Aggregator):
        """
        Interfaces with an `PyAutoFit` aggregator object to create instances of `Interferometer` objects from the results
        of a model-fit.

        The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

        - The interferometer visibilities data as a .fits file (`dataset/data.fits`).
        - The visibilities noise-map as a .fits file (`dataset/noise_map.fits`).
        - The uv wavelengths as a .fits file (`dataset/uv_wavelengths.fits`).
        - The real space mask defining the grid of the interferometer for the FFT (`dataset/real_space_mask.fits`).
        - The settings of the `Interferometer` data structure used in the fit (`dataset/settings.json`).

        The `aggregator` contains the path to each of these files, and they can be loaded individually. This class
        can load them all at once and create an `Interferometer` object via the `_interferometer_from` method.

        This class's methods returns generators which create the instances of the `Interferometer` objects. This ensures
        that large sets of results can be efficiently loaded from the hard-disk and do not require storing all
        `Interferometer` instances in the memory at once.

        For example, if the `aggregator` contains 3 model-fits, this class can be used to create a generator which
        creates instances of the corresponding 3 `Interferometer` objects.

        If multiple `Interferometer` objects were fitted simultaneously via analysis summing, the `fit.child_values()`
        method is instead used to load lists of the data, noise-map, PSF and mask and combine them into a list of
        `Interferometer` objects.

        This can be done manually, but this object provides a more concise API.

        Parameters
        ----------
        aggregator
            A `PyAutoFit` aggregator object which can load the results of model-fits.
        """
        self.aggregator = aggregator

    def dataset_gen_from(
        self,
        real_space_mask: Optional[aa.Mask2D] = None,
    ) -> List[aa.Interferometer]:
        """
        Returns a generator of `Interferometer` objects from an input aggregator.

        See `__init__` for a description of how the `Interferometer` objects are created by this method.

        Parameters
        ----------
        real_space_mask
            The real space mask.
        """
        func = partial(
            _interferometer_from,
            real_space_mask=real_space_mask,
        )

        return self.aggregator.map(func=func)



import numpy as np
from typing import List

import autoarray as aa

from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.galaxy.galaxies import Galaxies


class SimulatorInterferometer(aa.SimulatorInterferometer):
    def via_galaxies_from(self, galaxies: List[Galaxy], grid: aa.type.Grid2DLike):
        """
        Returns a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        image
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scales
            The scale of each pixel in arc seconds
        exposure_time_map
            An arrays representing the effective exposure time of each pixel.
        psf: PSF
            An arrays describing the PSF the simulated image is blurred with.
        add_poisson_noise: Bool
            If `True` poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        galaxies = Galaxies(galaxies=galaxies)

        image = galaxies.image_2d_from(grid=grid)

        return self.via_image_from(image=image)



import numpy as np
from scipy import interpolate
from typing import Tuple

from autoconf import cached_property

import autoarray as aa


class DatasetInterp:
    def __init__(self, dataset: aa.Imaging):
        """
        An ellipse interpolator, which contains a dataset (e.g. the image data and noise-map) and performs interpo.aiton
        calculations used for ellipse fitting.

        This object is used by the input to the `FitEllipse` object, which fits the dataset with ellipses and quantifies
        the goodness-of-fit via a residual map, likelihood, chi-squared and other quantities.

        The following quantities of the ellipse data are interpolated and used for the following tasks:

        - `data`: The image data, which shows the signal that is analysed and fitted with ellipses.

        - `noise_map`: The RMS standard deviation error in every pixel, which is used to compute the chi-squared value
        and likelihood of a fit.

        The `data` and `noise_map` are typically the same images of a galaxy used to perform standard light-profile
        fitting.

        Parameters
        ----------
        dataset
            The imaging data, containing the image data, noise map.
        """
        self.dataset = dataset

    @cached_property
    def points_interp(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        The points on which the interpolation from the 2D grid of data is performed.
        """

        x = self.dataset.mask.derive_grid.all_false.native[0, :, 1]
        y = np.flip(self.dataset.mask.derive_grid.all_false.native[:, 0, 0])

        return (x, y)

    @cached_property
    def data_interp(self) -> interpolate.RegularGridInterpolator:
        """
        Returns a 2D interpolation of the data, which is used to evaluate the data at any point in 2D space.
        """
        return interpolate.RegularGridInterpolator(
            points=self.points_interp,
            values=np.float64(self.dataset.data.native),
            bounds_error=False,
            fill_value=0.0,
        )

    @cached_property
    def noise_map_interp(self) -> interpolate.RegularGridInterpolator:
        """
        Returns a 2D interpolation of the noise-map, which is used to evaluate the noise-map at any point in 2D space.
        """

        return interpolate.RegularGridInterpolator(
            points=self.points_interp,
            values=np.float64(self.dataset.noise_map.native),
            bounds_error=False,
            fill_value=0.0,
        )



import numpy as np
from typing import Tuple

import autoarray as aa

from autogalaxy.profiles.mass.stellar.sersic import Sersic


class SersicCore(Sersic):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        effective_radius: float = 0.6,
        sersic_index: float = 4.0,
        radius_break: float = 0.01,
        intensity: float = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The elliptical cored-Sersic light profile.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        sersic_index
            Controls the concentration of the profile (lower -> less concentrated, higher -> more concentrated).
        radius_break
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity
            The intensity at the break radius.
        gamma
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """

        super().__init__(
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )

        self.radius_break = radius_break
        self.intensity = intensity
        self.alpha = alpha
        self.gamma = gamma

    def deflections_yx_2d_from(self, grid: aa.type.Grid2DLike, **kwargs):
        return self.deflections_2d_via_mge_from(grid=grid, **kwargs)

    def image_2d_via_radii_from(self, grid_radii: np.ndarray):
        """
        Calculate the intensity of the cored-Sersic light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        return np.multiply(
            np.multiply(
                self.intensity_prime,
                np.power(
                    np.add(
                        1,
                        np.power(np.divide(self.radius_break, grid_radii), self.alpha),
                    ),
                    (self.gamma / self.alpha),
                ),
            ),
            np.exp(
                np.multiply(
                    -self.sersic_constant,
                    (
                        np.power(
                            np.divide(
                                np.add(
                                    np.power(grid_radii, self.alpha),
                                    (self.radius_break**self.alpha),
                                ),
                                (self.effective_radius**self.alpha),
                            ),
                            (1.0 / (self.alpha * self.sersic_index)),
                        )
                    ),
                )
            ),
        )

    def decompose_convergence_via_mge(self):
        radii_min = self.effective_radius / 50.0
        radii_max = self.effective_radius * 20.0

        def core_sersic_2D(r):
            return (
                self.mass_to_light_ratio
                * self.intensity_prime
                * (1.0 + (self.radius_break / r) ** self.alpha)
                ** (self.gamma / self.alpha)
                * np.exp(
                    -self.sersic_constant
                    * (
                        (r**self.alpha + self.radius_break**self.alpha)
                        / self.effective_radius**self.alpha
                    )
                    ** (1.0 / (self.sersic_index * self.alpha))
                )
            )

        return self._decompose_convergence_via_mge(
            func=core_sersic_2D, radii_min=radii_min, radii_max=radii_max
        )

    @property
    def intensity_prime(self):
        """Overall intensity normalisation in the rescaled Core-Sersic light profiles (electrons per second)"""
        return (
            self.intensity
            * (2.0 ** (-self.gamma / self.alpha))
            * np.exp(
                self.sersic_constant
                * (
                    ((2.0 ** (1.0 / self.alpha)) * self.radius_break)
                    / self.effective_radius
                )
                ** (1.0 / self.sersic_index)
            )
        )


class SersicCoreSph(SersicCore):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        effective_radius: float = 0.6,
        sersic_index: float = 4.0,
        radius_break: float = 0.01,
        intensity: float = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
    ):
        """
        The elliptical cored-Sersic light profile.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        sersic_index
            Controls the concentration of the profile (lower -> less concentrated, higher -> more concentrated).
        radius_break
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity
            The intensity at the break radius.
        gamma
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super().__init__(
            centre=centre,
            ell_comps=(0.0, 0.0),
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            radius_break=radius_break,
            intensity=intensity,
            gamma=gamma,
            alpha=alpha,
        )
        self.radius_break = radius_break
        self.intensity = intensity
        self.alpha = alpha
        self.gamma = gamma



from astropy import units

import numpy as np
import warnings

from autogalaxy.cosmology.wrap import Planck15


def kappa_s_and_scale_radius_for_duffy(mass_at_200, redshift_object, redshift_source):
    """
    Computes the AutoGalaxy NFW parameters (kappa_s, scale_radius) for an NFW halo of the given
    mass, enforcing the Duffy '08 mass-concentration relation.

    Interprets mass as *`M_{200c}`*, not `M_{200m}`.
    """
    cosmology = Planck15()

    cosmic_average_density = (
        cosmology.critical_density(redshift_object).to(units.solMass / units.kpc**3)
    ).value

    critical_surface_density = (
        cosmology.critical_surface_density_between_redshifts_solar_mass_per_kpc2_from(
            redshift_0=redshift_object, redshift_1=redshift_source
        )
    )

    kpc_per_arcsec = cosmology.kpc_per_arcsec_from(redshift=redshift_object)

    radius_at_200 = (
        mass_at_200 / (200.0 * cosmic_average_density * (4.0 * np.pi / 3.0))
    ) ** (
        1.0 / 3.0
    )  # r200
    coefficient = 5.71 * (1.0 + redshift_object) ** (
        -0.47
    )  # The coefficient of Duffy mass-concentration (Duffy+2008)
    concentration = coefficient * (mass_at_200 / 2.952465309e12) ** (
        -0.084
    )  # mass-concentration relation. (Duffy+2008)
    de_c = (
        200.0
        / 3.0
        * (
            concentration**3
            / (np.log(1.0 + concentration) - concentration / (1.0 + concentration))
        )
    )  # rho_c

    scale_radius_kpc = radius_at_200 / concentration  # scale radius in kpc
    rho_s = cosmic_average_density * de_c  # rho_s
    kappa_s = rho_s * scale_radius_kpc / critical_surface_density  # kappa_s
    scale_radius = scale_radius_kpc / kpc_per_arcsec  # scale radius in arcsec

    return kappa_s, scale_radius, radius_at_200


def kappa_s_and_scale_radius_for_ludlow(
    mass_at_200, scatter_sigma, redshift_object, redshift_source
):
    """
    Computes the AutoGalaxy NFW parameters (kappa_s, scale_radius) for an NFW halo of the given
    mass, enforcing the Ludlow '16 mass-concentration relation.

    Interprets mass as *`M_{200c}`*, not `M_{200m}`.
    """
    from colossus.cosmology import cosmology as col_cosmology
    from colossus.halo.concentration import concentration as col_concentration

    warnings.filterwarnings("ignore")

    cosmology = Planck15()

    col_cosmo = col_cosmology.setCosmology("planck15")
    m_input = mass_at_200 * col_cosmo.h
    concentration = col_concentration(
        m_input, "200c", redshift_object, model="ludlow16"
    )

    concentration = 10.0 ** (np.log10(concentration) + scatter_sigma * 0.15)

    cosmic_average_density = (
        cosmology.critical_density(redshift_object).to(units.solMass / units.kpc**3)
    ).value

    critical_surface_density = (
        cosmology.critical_surface_density_between_redshifts_solar_mass_per_kpc2_from(
            redshift_0=redshift_object, redshift_1=redshift_source
        )
    )

    kpc_per_arcsec = cosmology.kpc_per_arcsec_from(redshift=redshift_object)

    radius_at_200 = (
        mass_at_200 / (200.0 * cosmic_average_density * (4.0 * np.pi / 3.0))
    ) ** (
        1.0 / 3.0
    )  # r200

    de_c = (
        200.0
        / 3.0
        * (
            concentration**3
            / (np.log(1.0 + concentration) - concentration / (1.0 + concentration))
        )
    )  # rho_c

    scale_radius_kpc = radius_at_200 / concentration  # scale radius in kpc
    rho_s = cosmic_average_density * de_c  # rho_s
    kappa_s = rho_s * scale_radius_kpc / critical_surface_density  # kappa_s
    scale_radius = scale_radius_kpc / kpc_per_arcsec  # scale radius in arcsec

    return kappa_s, scale_radius, radius_at_200



from typing import List

import autoarray as aa

from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.analysis.result import ResultDataset
from autogalaxy.imaging.fit_imaging import FitImaging


class ResultImaging(ResultDataset):
    @property
    def max_log_likelihood_fit(self) -> FitImaging:
        """
        An instance of a `FitImaging` corresponding to the maximum log likelihood model inferred by the non-linear
        search.
        """
        return self.analysis.fit_from(instance=self.instance)

    @property
    def max_log_likelihood_galaxies(self) -> List[Galaxy]:
        """
        An instance of galaxies corresponding to the maximum log likelihood model inferred by the non-linear search.

        The galaxies list is computed from the `max_log_likelihood_fit`, as this ensures that all linear light profiles
        are converted to normal light profiles with their `intensity` values updated.
        """
        return (
            self.max_log_likelihood_fit.model_obj_linear_light_profiles_to_light_profiles
        )

    @property
    def unmasked_model_image(self) -> aa.Array2D:
        """
        The model image of the maximum log likelihood model, creating without using a mask.
        """
        return self.max_log_likelihood_fit.unmasked_blurred_image

    @property
    def unmasked_model_image_of_galaxies(self) -> List[aa.Array2D]:
        """
        A list of the model image of every galaxy in the maximum log likelihood model, whereas all images are created
        without using a mask.
        """
        return self.max_log_likelihood_fit.unmasked_blurred_image_of_galaxies_list



import numpy as np
from typing import Tuple

from autogalaxy.cosmology.wrap import Planck15
from autogalaxy.profiles.mass.point.point import PointMass


class SMBH(PointMass):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        mass: float = 1e10,
        redshift_object: float = 0.5,
        redshift_source: float = 1.0,
    ):
        """
        Represents a supermassive black hole (SMBH).

        This uses the `PointMass` mass profile to represent the SMBH, where the SMBH mass in converted to the
        `PointMass` Einstein radius value.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        mass
            The mass of the SMBH in solar masses.
        redshift_object
            The redshift of the SMBH, which is used to convert its mass to an Einstein radius.
        redshift_source
            The redshift of the source galaxy, which is used to convert the mass of the SMBH to an Einstein radius.
        """

        self.mass = mass

        cosmology = Planck15()

        critical_surface_density = (
            cosmology.critical_surface_density_between_redshifts_from(
                redshift_0=redshift_object,
                redshift_1=redshift_source,
            )
        )
        mass_angular = mass / critical_surface_density
        einstein_radius = np.sqrt(mass_angular / np.pi)

        super().__init__(centre=centre, einstein_radius=einstein_radius)

        self.redshift_object = redshift_object
        self.redshift_source = redshift_source



from .isothermal import Isothermal, IsothermalSph
from .isothermal_core import IsothermalCore, IsothermalCoreSph
from .power_law import PowerLaw, PowerLawSph
from .power_law_broken import PowerLawBroken, PowerLawBrokenSph
from .power_law_core import PowerLawCore, PowerLawCoreSph
from .power_law_multipole import PowerLawMultipole



import os
from os import path
from typing import Dict, List, Union

from autoconf import conf
import autoarray as aa
import autoarray.plot as aplt

from autogalaxy.analysis.adapt_images.adapt_images import AdaptImages
from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.galaxy.galaxies import Galaxies
from autogalaxy.galaxy.plot.galaxy_plotters import GalaxyPlotter
from autogalaxy.galaxy.plot.galaxies_plotters import GalaxiesPlotter
from autogalaxy.galaxy.plot.adapt_plotters import AdaptPlotter

from autogalaxy.plot.include.two_d import Include2D
from autogalaxy.plot.mat_plot.one_d import MatPlot1D
from autogalaxy.plot.mat_plot.two_d import MatPlot2D


def setting(section: Union[List[str], str], name: str):
    if isinstance(section, str):
        return conf.instance["visualize"]["plots"][section][name]

    for sect in reversed(section):
        try:
            return conf.instance["visualize"]["plots"][sect][name]
        except KeyError:
            continue

    return conf.instance["visualize"]["plots"][section[0]][name]


def plot_setting(section: Union[List[str], str], name: str) -> bool:
    return setting(section, name)


class PlotterInterface:
    def __init__(self, image_path: str, title_prefix: str = None):
        """
        Provides an interface between an output path and all plotter objects.

        This is used to visualize the results of a model-fit, where the `image_path` points to the
        folder where the results of the model-fit are stored on your hard-disk, which is typically the `image` folder
        of a non-linear search.

        The `PlotterInterface` is typically used in the `Analysis` class of a non-linear search to visualize the maximum
        log likelihood model of the model-fit so far.

        The methods of the `PlotterInterface` are called throughout a non-linear search using the `Analysis.Visualizer`
        classes `visualize` method.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml`.

        Parameters
        ----------
        image_path
            The path on the hard-disk to the `image` folder of the non-linear searches results.
        title_prefix
            A string that is added before the title of all figures output by visualization, for example to
            put the name of the dataset and galaxy in the title.
        """
        self.image_path = image_path
        self.title_prefix = title_prefix

        self.include_2d = Include2D()

        os.makedirs(image_path, exist_ok=True)

    def mat_plot_1d_from(self, subfolders: str, format: str = "png") -> MatPlot1D:
        """
        Returns a 1D matplotlib plotting object whose `Output` class uses the `image_path`, such that it outputs
        images to the `image` folder of the non-linear search.

        Parameters
        ----------
        subfolders
            Subfolders between the `image` folder of the non-linear search and where the images are output. For example,
            images associsted with a fit are output to the subfolder `fit`.
        format
            The format images are output as, e.g. `.png` files.

        Returns
        -------
        MatPlot1D
            The 1D matplotlib plotter object.
        """
        return MatPlot1D(
            title=aplt.Title(prefix=self.title_prefix),
            output=aplt.Output(
                path=path.join(self.image_path, subfolders), format=format
            ),
        )

    def mat_plot_2d_from(self, subfolders, format="png") -> MatPlot2D:
        """
        Returns a 2D matplotlib plotting object whose `Output` class uses the `image_path`, such that it outputs
        images to the `image` folder of the non-linear search.

        Parameters
        ----------
        subfolders
            Subfolders between the `image` folder of the non-linear search and where the images are output. For example,
            images associsted with a fit are output to the subfolder `fit`.
        format
            The format images are output as, e.g. `.png` files.

        Returns
        -------
        MatPlot2D
            The 2D matplotlib plotter object.
        """
        return MatPlot2D(
            title=aplt.Title(prefix=self.title_prefix),
            output=aplt.Output(
                path=path.join(self.image_path, subfolders), format=format
            ),
        )

    def galaxies(
        self, galaxies: List[Galaxy], grid: aa.type.Grid2DLike, during_analysis: bool
    ):
        """
        Visualizes a list of galaxies.

        Images are output to the `image` folder of the `image_path` in a subfolder called `galaxies`. When
        used with a non-linear search the `image_path` points to the search's results folder and this function
        visualizes the maximum log likelihood galaxies inferred by the search so far.

        Visualization includes individual images of attributes of the galaxies (e.g. its image, convergence, deflection
        angles) and a subplot of all these attributes on the same figure.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml` under the
        [galaxies] header.

        Parameters
        ----------
        galaxies
            The maximum log likelihood galaxies of the non-linear search.
        grid
            A 2D grid of (y,x) arc-second coordinates used to perform ray-tracing, which is the masked grid tied to
            the dataset.
        during_analysis
            Whether visualization is performed during a non-linear search or once it is completed.
        """

        galaxies = Galaxies(galaxies=galaxies)

        def should_plot(name):
            return plot_setting(section="galaxies", name=name)

        subfolders = "galaxies"

        mat_plot_2d = self.mat_plot_2d_from(subfolders=subfolders)

        plotter = GalaxiesPlotter(
            galaxies=galaxies,
            grid=grid,
            mat_plot_2d=mat_plot_2d,
            include_2d=self.include_2d,
        )

        if should_plot("subplot_galaxy_images"):
            plotter.subplot_galaxy_images()

        plotter.figures_2d(
            image=should_plot("image"),
            convergence=should_plot("convergence"),
            potential=should_plot("potential"),
            deflections_y=should_plot("deflections"),
            deflections_x=should_plot("deflections"),
            magnification=should_plot("magnification"),
        )

        if not during_analysis and should_plot("all_at_end_png"):
            mat_plot_2d = self.mat_plot_2d_from(
                subfolders=path.join(subfolders, "end"),
            )

            plotter = GalaxiesPlotter(
                galaxies=galaxies,
                grid=grid,
                mat_plot_2d=mat_plot_2d,
                include_2d=self.include_2d,
            )

            plotter.figures_2d(
                image=True,
                convergence=True,
                potential=True,
                deflections_y=True,
                deflections_x=True,
                magnification=True,
            )

        mat_plot_2d = self.mat_plot_2d_from(subfolders="")

        plotter = GalaxiesPlotter(
            galaxies=galaxies,
            grid=grid,
            mat_plot_2d=mat_plot_2d,
            include_2d=self.include_2d,
        )

        if should_plot("subplot_galaxies"):
            plotter.subplot()

        if not during_analysis and should_plot("all_at_end_fits"):
            mat_plot_2d = self.mat_plot_2d_from(
                subfolders=path.join(subfolders, "fits"), format="fits"
            )

            plotter = GalaxiesPlotter(
                galaxies=galaxies,
                grid=grid,
                mat_plot_2d=mat_plot_2d,
                include_2d=self.include_2d,
            )

            plotter.figures_2d(
                image=True,
                convergence=True,
                potential=True,
                deflections_y=True,
                deflections_x=True,
                magnification=True,
            )

    def galaxies_1d(
        self, galaxies: [List[Galaxy]], grid: aa.type.Grid2DLike, during_analysis: bool
    ):
        """
        Visualizes a list of `Galaxy` objects.

        Images are output to the `image` folder of the `image_path` in a subfolder called `galaxies`. When
        used with a non-linear search the `image_path` points to the search's results folder and this function
        visualizes the maximum log likelihood `Galaxy`'s inferred by the search so far.

        Visualization includes individual images of attributes of each galaxy (e.g. 1D plots of their image,
        convergence) and a subplot of all these attributes on the same figure.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml` under the
        [galaxies] header.

        Parameters
        ----------
        galaxies
            A list of the maximum log likelihood `Galaxy`'s of the non-linear search.
        grid
            A 2D grid of (y,x) arc-second coordinates used to perform ray-tracing, which is the masked grid tied to
            the dataset.
        during_analysis
            Whether visualization is performed during a non-linear search or once it is completed.
        """

        def should_plot(name):
            return plot_setting(section="galaxies_1d", name=name)

        mat_plot_1d = self.mat_plot_1d_from(subfolders="galaxies_1d")

        for galaxy in galaxies:
            galaxy_plotter = GalaxyPlotter(
                galaxy=galaxy,
                grid=grid,
                mat_plot_1d=mat_plot_1d,
                include_2d=self.include_2d,
            )

            try:
                galaxy_plotter.figures_1d_decomposed(
                    image=should_plot("image"),
                    convergence=should_plot("convergence"),
                    potential=should_plot("potential"),
                )
            except OverflowError:
                pass

    def inversion(self, inversion: aa.Inversion, during_analysis: bool):
        """
        Visualizes an `Inversion` object.

        Images are output to the `image` folder of the `image_path` in a subfolder called `inversion`. When
        used with a non-linear search the `image_path` points to the search's results folder and this function
        visualizes the maximum log likelihood `Inversion` inferred by the search so far.

        Visualization includes individual images of attributes of the dataset (e.g. the reconstructed image, the
        reconstruction) and a subplot of all these attributes on the same figure.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml` under the
        [inversion] header.

        Parameters
        ----------
        inversion
            The inversion used to fit the dataset whose attributes are visualized.
        during_analysis
            Whether visualization is performed during a non-linear search or once it is completed.
        """

        def should_plot(name):
            return plot_setting(section="inversion", name=name)

        subfolders = "inversion"

        mat_plot_2d = self.mat_plot_2d_from(subfolders=subfolders)

        inversion_plotter = aplt.InversionPlotter(
            inversion=inversion, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        inversion_plotter.figures_2d(
            reconstructed_image=should_plot("reconstructed_image")
        )

        inversion_plotter.figures_2d_of_pixelization(
            pixelization_index=0,
            data_subtracted=should_plot("data_subtracted"),
            reconstructed_image=should_plot("reconstructed_image"),
            reconstruction=should_plot("reconstruction"),
            mesh_pixels_per_image_pixels=should_plot("mesh_pixels_per_image_pixels"),
            errors=should_plot("errors"),
            regularization_weights=should_plot("regularization_weights"),
        )

        mat_plot_2d = self.mat_plot_2d_from(subfolders="")

        inversion_plotter = aplt.InversionPlotter(
            inversion=inversion, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        if should_plot("subplot_inversion"):
            mapper_list = inversion.cls_list_from(cls=aa.AbstractMapper)

            for mapper_index in range(len(mapper_list)):
                inversion_plotter.subplot_of_mapper(mapper_index=mapper_index)

        if not during_analysis and should_plot("all_at_end_png"):
            mat_plot_2d = self.mat_plot_2d_from(subfolders=path.join(subfolders, "end"))

            inversion_plotter = aplt.InversionPlotter(
                inversion=inversion, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
            )

            inversion_plotter.figures_2d(reconstructed_image=True)

            inversion_plotter.figures_2d_of_pixelization(
                pixelization_index=0,
                reconstructed_image=True,
                reconstruction=True,
                errors=True,
                regularization_weights=True,
            )

        if not during_analysis and should_plot("all_at_end_fits"):
            mat_plot_2d = self.mat_plot_2d_from(
                subfolders=path.join(subfolders, "fits"), format="fits"
            )

            inversion_plotter = aplt.InversionPlotter(
                inversion=inversion, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
            )

            inversion_plotter.figures_2d(reconstructed_image=True)

            inversion_plotter.figures_2d_of_pixelization(
                pixelization_index=0,
                reconstructed_image=True,
                reconstruction=True,
                errors=True,
                regularization_weights=True,
                interpolate_to_uniform=True,
            )

    def adapt_images(
        self,
        adapt_images: AdaptImages,
    ):
        """
        Visualizes the adapt images used by a model-fit for adaptive pixelization mesh's and regularization.

        Images are output to the `image` folder of the `image_path` in a subfolder called `adapt`. When
        used with a non-linear search the `image_path` points to the search's results folder.

        Visualization includes an image of the overall adapt model image and a subplot of all galaxy images on the same
        figure.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml` under the
        [adapt] header.

        Parameters
        ----------
        adapt_images
            The adapt images (e.g. overall model image, individual galaxy images).
        """

        def should_plot(name):
            return plot_setting(section="adapt", name=name)

        mat_plot_2d = self.mat_plot_2d_from(subfolders="adapt")

        adapt_plotter = AdaptPlotter(
            mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        if should_plot("model_image"):
            adapt_plotter.figure_model_image(model_image=adapt_images.model_image)

        if should_plot("images_of_galaxies"):
            adapt_plotter.subplot_images_of_galaxies(
                adapt_galaxy_name_image_dict=adapt_images.galaxy_image_dict
            )



class Redshift(float):
    """
    Class used when assigning a redshift to a `Galaxy` object.

    This object is only required when making the `Redshift` of the `Galaxy` a free parameter in a model that
    is fitted.

    This is because **PyAutoFit** (which handles model-fitting), requires all parameters to be a Python class.

    The `Redshift` object does not need to be used for general **PyAutoGalaxy** use.

    Examples
    --------

    import autofit as af
    import autogalaxy as ag

    bulge = af.Model(ag.lp.Sersic)

    redshift = af.Model(ag.Redshift)
    redshift.redshift = af.UniformPrior(lower_limit=0.0, upper_limit=2.0)

    galaxy = af.Model(ag.Galaxy, redshift=redshift, bulge=bulge)

    model = af.Collection(galaxies=af.Collection(galaxy=galaxy))
    """

    def __new__(cls, redshift):
        # noinspection PyArgumentList
        return float.__new__(cls, redshift)

    def __init__(self, redshift):
        float.__init__(redshift)



import inspect
import numpy as np
from typing import ClassVar, Dict, List, Optional

from autoconf import cached_property
import autoarray as aa
import autofit as af

from autogalaxy.profiles.light.operated.abstract import (
    LightProfileOperated,
)

from autogalaxy.profiles.light.abstract import LightProfile

from autogalaxy import exc


class LightProfileLinear(LightProfile):
    """
    A linear light profile, which is a light profile whose `intensity` value is solved for via linear algebra
    using an inversion.

    Every standard light profile (e.g. `Serisic`, `Exponential`, etc.) has a linear light profile equivalent that
    behaves identically, except that the `intensity` is not explicitly set by the user but instead is inferred via
    a linear inversion.

    This means that when a linear light profile is used to perform a model-fit, it reduces the number of free parameters
    in the model-fit by 1, as the `intensity` parameter is inferred via the inversion.

    The `LightProfileLinear` class is an abstract class that should be used to make specific linear light profiles.
    This inheritance is used throughout the `galaxy.py`, `tracer.py` and other modules to extract linear light
    profiles when performing a fit to data.
    """

    @property
    def regularization(self):
        return None

    @property
    def _intensity(self):
        return 1.0

    @property
    def standard_lp_parent(self):
        """
        Returns the first parent class of the linear light profile which is not a linear light profile itself.

        The functions below (e.g. `lp_instance_from`) are used to convert a linear light profile to its parent
        standard light profile, where the input `intensity` is the value solved for in the linear inversion.

        This function is used to determine the parent class of the linear light profile that is used to create the
        standard light profile.

        This function also maps linear light and mass profiles to their standard light and mass profile counterparts.
        This specific mapping is required because their inheritence structure is different to the other light profiles.
        In the future, the ugly dictionary used to do this mapping should be removed for better code.

        Returns
        -------
        The parent class of the linear light profile that is not a linear light profile itself.
        """

        from autogalaxy.profiles import light_linear_and_mass_profiles as lmp_linear
        from autogalaxy.profiles import light_and_mass_profiles as lmp

        if isinstance(self, lmp.LightMassProfile):
            lmp_mapping_dict = {
                lmp_linear.Gaussian: lmp.Gaussian,
                lmp_linear.GaussianGradient: lmp.GaussianGradient,
                lmp_linear.SersicSph: lmp.SersicSph,
                lmp_linear.Sersic: lmp.Sersic,
                lmp_linear.SersicGradient: lmp.SersicGradient,
                lmp_linear.SersicGradientSph: lmp.SersicGradientSph,
                lmp_linear.ExponentialSph: lmp.ExponentialSph,
                lmp_linear.Exponential: lmp.Exponential,
                lmp_linear.ExponentialGradient: lmp.ExponentialGradient,
                lmp_linear.ExponentialGradientSph: lmp.ExponentialGradientSph,
                lmp_linear.DevVaucouleursSph: lmp.DevVaucouleursSph,
                lmp_linear.DevVaucouleurs: lmp.DevVaucouleurs,
                lmp_linear.SersicCoreSph: lmp.SersicCoreSph,
                lmp_linear.SersicCore: lmp.SersicCore,
            }

            return lmp_mapping_dict[self.__class__]

        bases = self.__class__.__bases__

        if self.__class__.__name__.endswith("Sph") or isinstance(
            self, LightProfileOperated
        ):
            bases = bases[0].__bases__

        for cls in bases:
            if not issubclass(cls, LightProfileLinear):
                return cls

    def parameters_dict_from(self, intensity: float) -> Dict[str, float]:
        """
        Returns a dictionary of the parameters of the linear light profile with the `intensity` added.

        This `intenisty` will likely have come from the value inferred via the linear inversion.

        Parameters
        ----------
        intensity
            Overall intensity normalisation of the not linear light profile that is created (units are dimensionless
            and derived from the data the light profile's image is compared too, which is expected to be electrons
            per second).
        """
        parameters_dict = vars(self)
        args = inspect.getfullargspec(self.standard_lp_parent).args
        args.remove("self")

        parameters_dict = {key: parameters_dict[key] for key in args}
        parameters_dict["intensity"] = intensity

        return parameters_dict

    def lp_instance_from(
        self, linear_light_profile_intensity_dict: Dict
    ) -> LightProfile:
        """
        Creates an instance of a linear light profile using its parent normal light profile (e.g. the non linear
        variant which has an `intensity` parameter).

        The `intensity` value of the profile created is passed into this function and used.

        Parameters
        ----------
        intensity
            Overall intensity normalisation of the not linear light profile that is created (units are dimensionless
            and derived from the data the light profile's image is compared too, which is expected to be electrons
            per second).
        """
        intensity = linear_light_profile_intensity_dict[self]
        parameters_dict = self.parameters_dict_from(intensity=intensity)

        return self.standard_lp_parent(**parameters_dict)


class LightProfileLinearObjFuncList(aa.AbstractLinearObjFuncList):
    def __init__(
        self,
        grid: aa.type.Grid1D2DLike,
        blurring_grid: aa.type.Grid1D2DLike,
        convolver: Optional[aa.Convolver],
        light_profile_list: List[LightProfileLinear],
        regularization=Optional[aa.reg.Regularization],
        run_time_dict: Optional[Dict] = None,
    ):
        """
        A list of linear light profiles which fits a dataset via linear algebra using the images of each linear light
        profile and the dataset.

        By inheriting from `AbstractLinearObjFuncList` this tells the parent project PyAutoArray that this class is a
        linear object that can be used in a linear inversion via its `Inversion` object. This means the linear light
        profiles are used to perform a linear inversion to fit the data, which solves for the `intensity` of each light profile.

        By overwriting the `mapping_matrix` function with a method that fills in each column of the matrix with the
        image of each linear light profile, this is then passed through the `inversion` package to perform the
        linear inversion. The API is identical to `Mapper` objects such that linear functions can easily be combined
        with mappers.

        The autoarray inversion module treats separely the `mapping_matrix` and the `operated_mapping_matrix`,
        where the latter is the mapping matrix after all additional operations (e.g. convolution, FFT) have been applied.
        The `operated_mapping_matrix` is used to perform the linear inversion. This class defines a function
        `operated_mapping_matrix_override` which overrides the `operated_mapping_matrix` calculation in autoarray
        to account for PSF convolution of the light profile images.

        This is a slightly weird design, is is required because although the autoarray inversion module has
        functionality to convolve ta `mapping_matrix` with a PSF in order to compute the `operated_mapping_matrix`,
        it does not have functionality to account for the fact that the light profile images have flux outside the
        masked region which is blurred into the masked region.

        For example, in `PyAutoGalaxy` and `PyAutoLens` the light of galaxies is represented using `LightProfile`
        objects, which describe the surface brightness of a galaxy as a function. By grouping lists of these linear
        light profiles into this class, autoarray solves for their intensity's linearly.

        Parameters
        ----------
        grid
            The (y,x) grid aligned with the data that is fitted, which is used to evaluate the image of each light
            profile.
        blurring_grid
            The blurring grid is all points whose light is outside the data's mask but close enough to the mask that
            it may be blurred into the mask. This is also used when evaluating the image of each light profile.
        convolver
            The convolver used to blur the light profile images of each light profile, the output of which
            makes up the columns of the `operated_mapping matrix`.
        light_profile_list
            A list of the linear light profiles that are used to fit the data via linear algebra.
        regularization
            The regularization scheme which may be applied to this linear object in order to smooth its solution.
        run_time_dict
            A dictionary which contains timing of certain functions calls which is used for profiling.
        """
        for light_profile in light_profile_list:
            if not isinstance(light_profile, LightProfileLinear):
                raise exc.ProfileException(
                    """
                    A light profile that is not a LightProfileLinear object has been input into the
                    LightProfileLinearObjFuncList object.

                    Only children of the LightProfileLinear class can be used in a linear inversion.
                    """
                )

        super().__init__(
            grid=grid, regularization=regularization, run_time_dict=run_time_dict
        )

        self.blurring_grid = blurring_grid
        self.convolver = convolver
        self.light_profile_list = light_profile_list

    @property
    def params(self) -> int:
        """
        The `params` property is used by the autoarray inversion module to track how many parameters are solved
        for in the linear inversion.

        This is used to define the dimensions of certain matrices used in the linear algebra, for example the
        `curvature_matrix` and `regularization_matrix`.

        For this class, the number of parameters is equal to the number of light profiles in the list,
        as each light profile contributes a single parameter to the inversion, its `intensity`.

        Returns
        -------
        The number of parameters that are solved for in the linear inversion.
        """
        return len(self.light_profile_list)

    @property
    def pixels_in_mask(self) -> int:
        """
        The number of pixels in the mask of the grid, which is used to define the dimensions of the `mapping_matrix`
        used in the linear inversion.

        This function has two ways to return the number of pixels in the mask, depending on whether the inversion
        is being performed on an imaging or interferometer dataset.

        Returns
        -------
        The number of pixels in the mask of the grid.
        """
        if isinstance(self.grid, aa.Grid2DOverSampled):
            return self.grid.pixels_in_mask
        return self.grid.mask.pixels_in_mask

    @property
    def mapping_matrix(self) -> np.ndarray:
        """
        Returns the `mapping_matrix` of the linear light profiles, where each column is the image of each light profile
        evaluated on the grid before operations are applied (e.g. convolution, FFT).

        This function iterates over each light profile in the list and evaluates its image on the grid, storing this
        image in the `mapping_matrix`.

        Returns
        -------
        The `mapping_matrix` of the linear light profiles.
        """
        mapping_matrix = np.zeros(shape=(self.pixels_in_mask, self.params))

        for pixel, light_profile in enumerate(self.light_profile_list):
            image_2d = light_profile.image_2d_from(grid=self.grid).slim

            mapping_matrix[:, pixel] = image_2d

        return mapping_matrix

    @cached_property
    def operated_mapping_matrix_override(self) -> Optional[np.ndarray]:
        """
        The inversion object takes the `mapping_matrix` of each linear object and combines it with the `Convolver`
        operator to perform a 2D convolution and compute the `operated_mapping_matrix`.

        If this property is overwritten this operation is not performed, with the `operated_mapping_matrix` output this
        property automatically used instead.

        This is used for a linear light profile because the in-built mapping matrix convolution does not account for
        how light profile images have flux outside the masked region which is blurred into the masked region. This
        flux is outside the region that defines the `mapping_matrix` and thus this override is required to properly
        incorporate it.

        Returns
        -------
        A blurred mapping matrix of dimensions (total_mask_pixels, 1) which overrides the mapping matrix calculations
        performed in the linear equation solvers.
        """

        if isinstance(self.light_profile_list[0], LightProfileOperated):
            return self.mapping_matrix

        operated_mapping_matrix = np.zeros(shape=(self.pixels_in_mask, self.params))

        for pixel, light_profile in enumerate(self.light_profile_list):
            image_2d = light_profile.image_2d_from(grid=self.grid)

            blurring_image_2d = light_profile.image_2d_from(grid=self.blurring_grid)

            blurred_image_2d = self.convolver.convolve_image(
                image=image_2d, blurring_image=blurring_image_2d
            )

            operated_mapping_matrix[:, pixel] = blurred_image_2d

        return operated_mapping_matrix



import numpy as np
from typing import List, Optional

from autoconf import cached_property

import autoarray as aa

from autogalaxy.ellipse.dataset_interp import DatasetInterp
from autogalaxy.ellipse.ellipse.ellipse import Ellipse
from autogalaxy.ellipse.ellipse.ellipse_multipole import EllipseMultipole


class FitEllipse(aa.FitDataset):
    def __init__(
        self,
        dataset: aa.Imaging,
        ellipse: Ellipse,
        multipole_list: Optional[List[EllipseMultipole]] = None,
    ):
        """
        A fit to a `DatasetInterp` dataset, using a model image to represent the observed data and noise-map.

        Parameters
        ----------
        dataset
            The dataset containing the signal and noise-map that is fitted.

        """
        super().__init__(dataset=dataset)

        self.ellipse = ellipse
        self.multipole_list = multipole_list

    @cached_property
    def interp(self) -> DatasetInterp:
        """
        Returns a class which handles the interpolation of values from the image data and noise-map, so that they
        can be mapped to each ellipse for the fit.
        """
        return DatasetInterp(dataset=self.dataset)

    def points_from_major_axis_from(self, flip_y: bool = False) -> np.ndarray:
        """
        Returns the (y,x) coordinates on the ellipse that are used to interpolate the data and noise-map values.

        These points are computed by overlaying the ellipse over the 2D data and noise-map and computing the (y,x)
        coordinates on the ellipse that are closest to the data points.

        If multipole components are used, the points are also perturbed by the multipole components.

        When plotting the ellipses, y coordinates must be flipped to match the convention of the y-axis increasing
        downwards in 2D data, which is performed by setting `flip_y=True`.

        Returns
        -------
        The (y,x) coordinates on the ellipse where the interpolation occurs.
        """
        points = self.ellipse.points_from_major_axis_from(
            pixel_scale=self.dataset.pixel_scales[0], flip_y=flip_y
        )

        if self.multipole_list is not None:
            for multipole in self.multipole_list:
                points = multipole.points_perturbed_from(
                    pixel_scale=self.dataset.pixel_scales[0],
                    points=points,
                    ellipse=self.ellipse,
                )

        return points

    @cached_property
    def _points_from_major_axis(self) -> np.ndarray:
        """
        Returns cached (y,x) coordinates on the ellipse that are used to interpolate the data and noise-map values.

        Returns
        -------
        The (y,x) coordinates on the ellipse where the interpolation occurs.
        """
        return self.points_from_major_axis_from()

    @property
    def data_interp(self) -> aa.ArrayIrregular:
        """
        Returns the data values of the dataset that the ellipse fits, which are computed by overlaying the ellipse over
        the 2D data and performing a 2D interpolation at discrete (y,x) coordinates on the ellipse.

        The (y,x) coordinates on the ellipse where the interpolation occurs are computed in the
        `points_from_major_axis` property of the `Ellipse` class, with the documentation describing how these points
        are computed.

        Returns
        -------
        The data values of the ellipse fits, computed via a 2D interpolation of where the ellipse
        overlaps the data.
        """

        return aa.ArrayIrregular(
            values=self.interp.data_interp(self._points_from_major_axis)
        )

    @property
    def noise_map_interp(self) -> aa.ArrayIrregular:
        """
        Returns the noise-map values of the dataset that the ellipse fits, which are computed by overlaying the ellipse
        over the 2D noise-map and performing a 2D interpolation at discrete (y,x) coordinates on the ellipse.

        The (y,x) coordinates on the ellipse where the interpolation occurs are computed in the
        `points_from_major_axis` property of the `Ellipse` class, with the documentation describing how these points
        are computed.

        Returns
        -------
        The noise-map values of the ellipse fits, computed via a 2D interpolation of where the ellipse
        overlaps the noise-map.
        """
        return aa.ArrayIrregular(
            values=self.interp.noise_map_interp(self._points_from_major_axis)
        )

    @property
    def signal_to_noise_map_interp(self) -> aa.ArrayIrregular:
        """
        Returns the signal-to-noise-map of the dataset that the ellipse fits, which is computed by overlaying the ellipse
        over the 2D data and noise-map and performing a 2D interpolation at discrete (y,x) coordinates on the ellipse.

        Returns
        -------
        The signal-to-noise-map values of the ellipse fits, computed via a 2D interpolation of where
        the ellipse overlaps the data and noise-map.
        """
        return aa.ArrayIrregular(values=self.data_interp / self.noise_map_interp)

    @property
    def model_data(self) -> aa.ArrayIrregular:
        """
        Returns the model-data, which is the data values where the ellipse overlaps the data minus the mean
        value of these data values.

        By subtracting the mean of the data values from each data value, the model data quantifies how often there
        are large variations in the data values over the ellipse.

        For example, if every data value subtended by the ellipse are close to one another, the difference between
        the data values and the mean will be small.

        Conversely, if some data values are much higher or lower than the mean, the model data will be large.

        Returns
        -------
        The model data values of the ellipse fit, which are the data values minus the mean of the data values.
        """
        return self.data_interp

    @property
    def residual_map(self):
        """
        Returns the residual-map of the fit, which is the data minus the model data and therefore the same
        as the model data.

        Returns
        -------
        The residual-map of the fit, which is the data minus the model data and therefore the same as the model data.
        """
        return aa.ArrayIrregular(values=self.model_data - np.nanmean(self.model_data))

    @property
    def normalized_residual_map(self) -> aa.ArrayIrregular:
        """
        Returns the normalized residual-map of the fit, which is the residual-map divided by the noise-map.

        The residual map and noise map are computed by overlaying the ellipse over the 2D data and noise-map and
        performing a 2D interpolation at discrete (y,x) coordinates on the ellipse. See the documentation of the
        `residual_map` and `noise_map` properties for more details.

        Returns
        -------
        The normalized residual-map of the fit, which is the residual-map divided by the noise-map.
        """

        normalized_residual_map = (self.residual_map) / self.noise_map_interp

        # NOTE:
        idx = np.logical_or(
            np.isnan(normalized_residual_map), np.isinf(normalized_residual_map)
        )
        normalized_residual_map[idx] = 0.0

        return aa.ArrayIrregular(values=normalized_residual_map)

    @property
    def chi_squared_map(self) -> aa.ArrayIrregular:
        """
        Returns the chi-squared-map of the fit, which is the normalized residual-map squared.

        The normalized residual map is computed by overlaying the ellipse over the 2D data and noise-map and
        performing a 2D interpolation at discrete (y,x) coordinates on the ellipse. See the documentation of the
        `normalized_residual_map` property for more details.

        Returns
        -------
        The chi-squared-map of the fit, which is the normalized residual-map squared.
        """
        return aa.ArrayIrregular(values=self.normalized_residual_map**2.0)

    @property
    def chi_squared(self) -> float:
        """
        The sum of the chi-squared-map, which quantifies how well the model data represents the data and noise-map.

        The chi-squared-map is computed by overlaying the ellipse over the 2D data and noise-map and
        performing a 2D interpolation at discrete (y,x) coordinates on the ellipse. See the documentation of the
        `chi_squared_map` property for more details.

        Returns
        -------
        The chi-squared of the fit.
        """
        return float(np.sum(self.chi_squared_map))

    @property
    def noise_normalization(self):
        """
        The noise normalization term of the log likelihood, which is the sum of the log noise-map values squared.

        Returns
        -------
        The noise normalization term of the log likelihood.
        """
        return np.sum(np.log(2 * np.pi * self.noise_map_interp**2.0))

    @property
    def log_likelihood(self):
        """
        The log likelihood of the fit, which quantifies how well the model data represents the data and noise-map.

        Returns
        -------
        The log likelihood of the fit.
        """
        return -0.5 * (self.chi_squared)

    @property
    def figure_of_merit(self) -> float:
        """
        The figure of merit of the fit, which is passed by the `Analysis` class to the non-linear search to
        determine the best-fit solution.

        Returns
        -------
        The figure of merit of the fit.
        """
        return self.log_likelihood



from .chameleon import Chameleon, ChameleonSph
from .dev_vaucouleurs import DevVaucouleurs, DevVaucouleursSph
from .exponential import Exponential, ExponentialSph
from .gaussian import Gaussian
from .gaussian_gradient import GaussianGradient
from .sersic import Sersic, SersicSph
from .sersic_core import SersicCore, SersicCoreSph
from .sersic_gradient import SersicGradient, SersicGradientSph



from typing import Tuple

from autogalaxy.profiles.mass.dark.nfw import NFWSph

from astropy import units

import numpy as np
import warnings

from autogalaxy.cosmology.wrap import Planck15


def kappa_s_and_scale_radius(
    virial_mass, concentration, virial_overdens, redshift_object, redshift_source
):
    cosmology = Planck15()

    cosmic_average_density = (
        cosmology.critical_density(redshift_object).to(units.solMass / units.kpc**3)
    ).value

    critical_surface_density = (
        cosmology.critical_surface_density_between_redshifts_solar_mass_per_kpc2_from(
            redshift_0=redshift_object, redshift_1=redshift_source
        )
    )

    kpc_per_arcsec = cosmology.kpc_per_arcsec_from(redshift=redshift_object)

    virial_radius = (
        virial_mass / (virial_overdens * cosmic_average_density * (4.0 * np.pi / 3.0))
    ) ** (
        1.0 / 3.0
    )  # r_vir
    de_c = (
        virial_overdens
        / 3.0
        * (
            concentration**3
            / (np.log(1.0 + concentration) - concentration / (1.0 + concentration))
        )
    )  # rho_c

    scale_radius_kpc = virial_radius / concentration  # scale radius in kpc
    rho_s = cosmic_average_density * de_c  # rho_s
    kappa_s = rho_s * scale_radius_kpc / critical_surface_density  # kappa_s
    scale_radius = scale_radius_kpc / kpc_per_arcsec  # scale radius in arcsec

    return kappa_s, scale_radius, virial_radius


class NFWVirialMassConcSph(NFWSph):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        virial_mass: float = 1e12,
        concentration: float = 10,
        virial_overdens: float = 200,
        redshift_object: float = 0.5,
        redshift_source: float = 1.0,
    ):
        """
        Spherical NFW profile initialized with the mass and concentration of the halo.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        virial_mass
            The virial mass of the dark matter halo.
        concentration
            The concentration of the dark matter halo.
        virial_overdens
            The virial overdensity.
        """
        self.virial_mass = virial_mass
        self.concentration = concentration
        self.virial_overdens = virial_overdens

        (
            kappa_s,
            scale_radius,
            virial_radius,
        ) = kappa_s_and_scale_radius(
            virial_mass=virial_mass,
            concentration=concentration,
            virial_overdens=virial_overdens,
            redshift_object=redshift_object,
            redshift_source=redshift_source,
        )

        super().__init__(centre=centre, kappa_s=kappa_s, scale_radius=scale_radius)



from os import path

import autoarray as aa
import autoarray.plot as aplt

from autogalaxy.imaging.fit_imaging import FitImaging
from autogalaxy.imaging.plot.fit_imaging_plotters import FitImagingPlotter
from autogalaxy.analysis.plotter_interface import PlotterInterface

from autogalaxy.analysis.plotter_interface import plot_setting


class PlotterInterfaceImaging(PlotterInterface):
    def imaging(self, dataset: aa.Imaging):
        """
        Visualizes an `Imaging` dataset object.

        Images are output to the `image` folder of the `image_path` in a subfolder called `imaging`. When used with
        a non-linear search the `image_path` points to the search's results folder.
        `.
        Visualization includes individual images of attributes of the dataset (e.g. the image, noise map, PSF) and a
        subplot of all these attributes on the same figure.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml` under the
        [dataset] header.

        Parameters
        ----------
        dataset
            The imaging dataset whose attributes are visualized.
        """

        def should_plot(name):
            return plot_setting(section=["dataset", "imaging"], name=name)

        mat_plot_2d = self.mat_plot_2d_from(subfolders="dataset")

        dataset_plotter = aplt.ImagingPlotter(
            dataset=dataset, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        dataset_plotter.figures_2d(
            data=should_plot("data"),
            noise_map=should_plot("noise_map"),
            psf=should_plot("psf"),
            signal_to_noise_map=should_plot("signal_to_noise_map"),
            over_sampling=should_plot("over_sampling"),
            over_sampling_non_uniform=should_plot("over_sampling_non_uniform"),
            over_sampling_pixelization=should_plot("over_sampling_pixelization"),
        )

        mat_plot_2d = self.mat_plot_2d_from(subfolders="")

        dataset_plotter = aplt.ImagingPlotter(
            dataset=dataset, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        if should_plot("subplot_dataset"):
            dataset_plotter.subplot_dataset()

    def fit_imaging(
        self, fit: FitImaging, during_analysis: bool, subfolders: str = "fit_dataset"
    ):
        """
        Visualizes a `FitImaging` object, which fits an imaging dataset.

        Images are output to the `image` folder of the `image_path` in a subfolder called `fit`. When
        used with a non-linear search the `image_path` points to the search's results folder and this function
        visualizes the maximum log likelihood `FitImaging` inferred by the search so far.

        Visualization includes individual images of attributes of the `FitImaging` (e.g. the model data, residual map)
        and a subplot of all `FitImaging`'s images on the same figure.

        The images output by the `PlotterInterface` are customized using the file `config/visualize/plots.yaml` under the
        [fit] header.

        Parameters
        ----------
        fit
            The maximum log likelihood `FitImaging` of the non-linear search which is used to plot the fit.
        during_analysis
            Whether visualization is performed during a non-linear search or once it is completed.
        visuals_2d
            An object containing attributes which may be plotted over the figure (e.g. the centres of mass and light
            profiles).
        """

        def should_plot(name):
            return plot_setting(section=["fit", "fit_imaging"], name=name)

        mat_plot_2d = self.mat_plot_2d_from(subfolders=subfolders)

        fit_plotter = FitImagingPlotter(
            fit=fit, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        fit_plotter.figures_2d(
            data=should_plot("data"),
            noise_map=should_plot("noise_map"),
            signal_to_noise_map=should_plot("signal_to_noise_map"),
            model_image=should_plot("model_data"),
            residual_map=should_plot("residual_map"),
            normalized_residual_map=should_plot("normalized_residual_map"),
            chi_squared_map=should_plot("chi_squared_map"),
        )

        fit_plotter.figures_2d_of_galaxies(
            subtracted_image=should_plot("subtracted_images_of_galaxies"),
            model_image=should_plot("model_images_of_galaxies"),
        )

        if should_plot("subplot_fit"):
            fit_plotter.subplot_fit()

        if should_plot("subplot_of_galaxies"):
            fit_plotter.subplot_of_galaxies()

        if not during_analysis and should_plot("all_at_end_png"):
            mat_plot_2d = self.mat_plot_2d_from(subfolders=path.join(subfolders, "end"))

            fit_plotter = FitImagingPlotter(
                fit=fit, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
            )

            fit_plotter.figures_2d(
                data=True,
                noise_map=True,
                signal_to_noise_map=True,
                model_image=True,
                residual_map=True,
                normalized_residual_map=True,
                chi_squared_map=True,
            )

            fit_plotter.figures_2d_of_galaxies(subtracted_image=True, model_image=True)

        if not during_analysis and should_plot("all_at_end_fits"):
            mat_plot_2d = self.mat_plot_2d_from(
                subfolders=path.join(subfolders, "fits"), format="fits"
            )

            fit_plotter = FitImagingPlotter(
                fit=fit, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
            )

            fit_plotter.figures_2d(
                data=True,
                noise_map=True,
                signal_to_noise_map=True,
                model_image=True,
                residual_map=True,
                normalized_residual_map=True,
                chi_squared_map=True,
            )

            fit_plotter.figures_2d_of_galaxies(
                subtracted_image=True,
                model_image=True,
            )



from typing import Tuple

from autogalaxy.profiles.light_and_mass_profiles import LightMassProfile

from autogalaxy.profiles.light import linear as lp_linear
from autogalaxy.profiles import mass as mp


class Gaussian(lp_linear.Gaussian, mp.Gaussian, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        sigma: float = 1.0,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The elliptical Gaussian light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles.
        """
        lp_linear.Gaussian.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            sigma=sigma,
        )
        mp.Gaussian.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            sigma=sigma,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class GaussianGradient(lp_linear.Gaussian, mp.GaussianGradient, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        sigma: float = 1.0,
        mass_to_light_ratio_base: float = 1.0,
        mass_to_light_gradient: float = 0.0,
        mass_to_light_radius: float = 1.0,
    ):
        """
        The elliptical Gaussian light profile with a gradient in its mass to light conversion.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio_base
            The base mass-to-light ratio of the profile, which is the mass-to-light ratio of the Gaussian before it
            is scaled by values that adjust its mass-to-light ratio based on the reference radius and gradient.
        mass_to_light_gradient
            The mass-to-light radial gradient of the profile, whereby positive values means there is more mass
            per unit light within the reference radius.
        mass_to_light_radius
            The radius where the mass-to-light ratio is equal to the base mass-to-light ratio, such that there will be
            more of less mass per unit light within this radius depending on the mass-to-light gradient.
        """
        lp_linear.Gaussian.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            sigma=sigma,
        )
        mp.GaussianGradient.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            sigma=sigma,
            mass_to_light_ratio_base=mass_to_light_ratio_base,
            mass_to_light_gradient=mass_to_light_gradient,
            mass_to_light_radius=mass_to_light_radius,
        )


class Sersic(lp_linear.Sersic, mp.Sersic, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The elliptical Sersic light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles.
        """
        lp_linear.Sersic.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
        )
        mp.Sersic.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SersicSph(Sersic, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The spherical Sersic light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre..
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles.
        """
        Sersic.__init__(
            self,
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class Exponential(Sersic, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The elliptical Exponential light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        """
        Sersic.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class ExponentialSph(Exponential, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The spherical Exponential light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        """
        Exponential.__init__(
            self,
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class DevVaucouleurs(Sersic, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The elliptical Dev Vaucouleurs light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        """
        super().__init__(
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=4.0,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class DevVaucouleursSph(DevVaucouleurs, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The spherical Dev Vaucouleurs light and mass profile.

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        """
        DevVaucouleurs.__init__(
            self,
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SersicGradient(lp_linear.Sersic, mp.SersicGradient, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        The elliptical Sersic light and mass profile, with a radial gradient in the conversion of light to mass..

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre..
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        sersic_index
            The concentration of the light profiles.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles.
        mass_to_light_gradient
            The mass-to-light radial gradient.
        """
        lp_linear.Sersic.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
        )
        mp.SersicGradient.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )


class SersicGradientSph(SersicGradient, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        The spherical Sersic light and mass profile, with a radial gradient in the conversion of light to mass..

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        sersic_index
            The concentration of the light profiles
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        mass_to_light_gradient
            The mass-to-light radial gradient.
        """

        SersicGradient.__init__(
            self,
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )


class ExponentialGradient(SersicGradient, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        The elliptical Exponential light and mass profile, with a radial gradient in the conversion of light to mass..

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        mass_to_light_gradient
            The mass-to-light radial gradient.
        """

        SersicGradient.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )


class ExponentialGradientSph(SersicGradientSph, LightMassProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        The spherical Exponential light and mass profile, with a radial gradient in the conversion of light to mass..

        This simultaneously represents the luminous emission and stellar mass of a galaxy.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        mass_to_light_gradient
            The mass-to-light radial gradient.
        """

        SersicGradientSph.__init__(
            self,
            centre=centre,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )


class SersicCore(lp_linear.SersicCore, mp.SersicCore, LightMassProfile):
    """
    The elliptical cored-Sersic light and mass profile.

    This simultaneously represents the luminous emission and stellar mass of a galaxy.

    Parameters
    ----------
    centre
        The grid of The (y,x) arc-second coordinates of the profile centre.
    ell_comps
        The first and second ellipticity components of the elliptical coordinate system.
    intensity
        Overall flux intensity normalisation in the light profiles (electrons per second).
    effective_radius
        The radius containing half the light of this light profile.
    radius_break
        The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
    gamma
        The logarithmic power-law slope of the inner core profiles
    alpha
        Controls the sharpness of the transition between the inner core / outer Sersic profiles.
    mass_to_light_ratio
        The mass-to-light ratio of the light profiles.
    """

    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        effective_radius: float = 0.6,
        sersic_index: float = 4.0,
        radius_break: float = 0.01,
        intensity: float = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
        mass_to_light_ratio: float = 1.0,
    ):
        lp_linear.SersicCore.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            radius_break=radius_break,
            gamma=gamma,
            alpha=alpha,
        )
        mp.SersicCore.__init__(
            self,
            centre=centre,
            ell_comps=ell_comps,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            radius_break=radius_break,
            intensity=intensity,
            gamma=gamma,
            alpha=alpha,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SersicCoreSph(SersicCore, LightMassProfile):
    """
    The spherical cored-Sersic light and mass profile.

    This simultaneously represents the luminous emission and stellar mass of a galaxy.

    Parameters
    ----------
    centre
        The grid of The (y,x) arc-second coordinates of the profile centre.
    ell_comps
        The first and second ellipticity components of the elliptical coordinate system.
    intensity
        Overall flux intensity normalisation in the light profiles (electrons per second).
    effective_radius
        The radius containing half the light of this light profile.
    radius_break
        The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
    gamma
        The logarithmic power-law slope of the inner core profiles
    alpha
        Controls the sharpness of the transition between the inner core / outer Sersic profiles.
    mass_to_light_ratio
        The mass-to-light ratio of the light profiles.
    """

    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        effective_radius: float = 0.6,
        sersic_index: float = 4.0,
        radius_break: float = 0.01,
        intensity: float = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The SersicSph mass profile, the mass profiles of the light profiles that are used to fit_normal and
        subtract the lens model_galaxy's light.

        Parameters
        ----------
        centre
            The grid of The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius
            The radius containing half the light of this light profile.
        mass_to_light_ratio
            The mass-to-light ratio of the light profiles
        """
        SersicCore.__init__(
            self,
            centre=centre,
            ell_comps=(0.0, 0.0),
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            radius_break=radius_break,
            intensity=intensity,
            gamma=gamma,
            alpha=alpha,
            mass_to_light_ratio=mass_to_light_ratio,
        )



from .gaussian import Gaussian, GaussianSph
from .moffat import Moffat, MoffatSph
from .sersic import (
    Sersic,
    SersicSph,
)
from .exponential import Exponential, ExponentialSph
from .dev_vaucouleurs import DevVaucouleurs, DevVaucouleursSph
from .sersic_core import SersicCore, SersicCoreSph
from .exponential_core import ExponentialCore, ExponentialCoreSph
from .chameleon import Chameleon, ChameleonSph
from .eff import (
    ElsonFreeFall,
    ElsonFreeFallSph,
)
from .shapelets.polar import ShapeletPolarSph, ShapeletPolar
from .shapelets.cartesian import ShapeletCartesianSph, ShapeletCartesian
from .shapelets.exponential import ShapeletExponentialSph, ShapeletExponential



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from os import path

from autoconf import conf
from autoconf.dictable import from_json, output_to_json

import autoarray as aa
import autofit as af

if TYPE_CHECKING:
    from autogalaxy.analysis.result import Result

from autogalaxy.profiles.light.abstract import LightProfile
from autogalaxy.profiles.mass.abstract.abstract import MassProfile
from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.profiles.basis import Basis

from autogalaxy.profiles.light import standard as lp
from autogalaxy.profiles.light import linear as lp_linear
from autogalaxy.profiles import light_and_mass_profiles as lmp
from autogalaxy.profiles import light_linear_and_mass_profiles as lmp_linear


def mass_from(mass, mass_result, unfix_mass_centre: bool = False) -> af.Model:
    """
    Returns an updated mass `Model` whose priors are initialized from a previous results in a pipeline.

    It includes an option to unfix the input `mass_centre` used previously (e.g. in the SLaM SOURCE PIPELINE), such
    that if the `mass_centre` were fixed (e.g. to (0.0", 0.0")) it becomes a free parameter.

    This function generically links any `MassProfile` to any `MassProfile`, pairing parameters which share the
    same path.

    Parameters
    ----------
    mass
        The mass profile about to be fitted, whose priors are updated based on the previous results.
    mass_result
        The mass profile inferred as a result of the previous pipeline, whose priors are used to update the
        input mass.
    unfix_mass_centre
        If the `mass_centre` was fixed to an input value in a previous pipeline, then `True` will unfix it and make it
        free parameters that are fitted for.

    Returns
    -------
    af.Model(mp.MassProfile)
        The total mass profile whose priors are initialized from a previous result.
    """

    mass.take_attributes(source=mass_result)

    if unfix_mass_centre and isinstance(mass.centre, tuple):
        centre_tuple = mass.centre

        mass.centre = af.Model(mass.cls).centre

        mass.centre.centre_0 = af.GaussianPrior(mean=centre_tuple[0], sigma=0.05)
        mass.centre.centre_1 = af.GaussianPrior(mean=centre_tuple[1], sigma=0.05)

    return mass


def source_custom_model_from(result: Result, source_is_model: bool = False) -> af.Model:
    """
    Setup the source model using the previous pipeline's source result.

    The source light model is not specified by the MASS PIPELINE and the previous SOURCE PIPELINE is used to
    determine whether the source model is parametric or an inversion.

    The source can be returned as an `instance` or `model`, depending on the optional input. The default behaviour is
    to return parametric sources as a model (give they must be updated to properly compute a new mass model) and
    return inversions as an instance (as they have sufficient flexibility to typically not required updating).

    Parameters
    ----------
    result
        The result of the previous source pipeline.
    source_is_model
        If `True` the source is returned as a *model* where the parameters are fitted for using priors of the
        search result it is loaded from. If `False`, it is an instance of that search's result.

    Returns
    -------
    af.Model(Galaxy)
        The source galaxy with its light profile(s) and pixelization, set up as a model or instance according to
        the input `source_is_model`.
    """

    redshift = result.instance.galaxies.source.redshift

    if not hasattr(result.instance.galaxies.source, "pixelization"):
        if source_is_model:
            return af.Model(
                Galaxy,
                redshift=redshift,
                bulge=result.model.galaxies.source.bulge,
                disk=result.model.galaxies.source.disk,
            )

        return af.Model(
            Galaxy,
            redshift=redshift,
            bulge=result.instance.galaxies.source.bulge,
            disk=result.instance.galaxies.source.disk,
        )

    if source_is_model:
        pixelization = af.Model(
            aa.Pixelization,
            image_mesh=result.instance.galaxies.source.pixelization.image_mesh,
            mesh=result.instance.galaxies.source.pixelization.mesh,
            regularization=result.model.galaxies.source.pixelization.regularization,
        )

        return af.Model(
            Galaxy,
            redshift=redshift,
            pixelization=pixelization,
        )

    pixelization = af.Model(
        aa.Pixelization,
        image_mesh=result.instance.galaxies.source.pixelization.image_mesh,
        mesh=result.instance.galaxies.source.pixelization.mesh,
        regularization=result.instance.galaxies.source.pixelization.regularization,
    )

    return af.Model(
        Galaxy,
        redshift=redshift,
        pixelization=pixelization,
    )


def source_from(
    result: Result,
) -> af.Model:
    """
    Setup the source model for a MASS PIPELINE using the previous SOURCE PIPELINE results.

    The source light model is not specified by the  MASS PIPELINE and the previous SOURCE PIPELINE is used to
    determine whether the source model is parametric or an inversion.

    The source is returned as a model if it is parametric (given its parameters must be fitted for to properly compute
    a new mass model) whereas inversions are returned as an instance (as they have sufficient flexibility to not
    require updating). This behaviour can be customized in SLaM pipelines by replacing this method with the
    `source_from` method.

    Parameters
    ----------
    result
        The result of the previous source pipeline.

    Returns
    -------
    af.Model(Galaxy)
        The source galaxy with its light profile(s) and pixelization, set up as a model or instance according to
        the input `result`.
    """

    if hasattr(result.instance.galaxies.source, "pixelization"):
        if result.instance.galaxies.source.pixelization is not None:
            return source_custom_model_from(result=result, source_is_model=False)
    return source_custom_model_from(result=result, source_is_model=True)


def clumps_from(
    result: Result,
    light_as_model: bool = False,
    mass_as_model: bool = False,
    free_centre: bool = False,
) -> af.Collection:
    """
    The clump API models the light and / or mass of additional galaxies surrouding the main galaxy or strong lens
    system.

    This function performs model composition of clumps for fits using the search chaining API. It makes it possible to
    pass the clump parameters from a previous search to a new search, such that the new clumps are either treated
    as an instance or model component.

    This function currently requires that mass profiles are `IsothermalSph` objects and that light profiles are
    `Sersic` objects. This will be generalised in the future.

    Parameters
    ----------
    result
        The result, which includes clumps, of the previous search, which via prior passing are used to create the new
        clump model.
    light_as_model
        If `True`, the clump light is passed as a model component, else it is a fixed instance.
    mass_as_model
        If `True`, the clump mass is passed as a model component, else it is a fixed instance.
    free_centre
        If `True`, the requested mass and/or light model centres are passed as a model, else they are fixed.

    Returns
    -------
    af.Collection
        A collection of clump `Galaxy` objects, where each clump is either an instance or model component.
    """
    # ideal API:

    # clumps = result.instance.clumps.as_model((LightProfile, mp.MassProfile,), fixed="centre", prior_pass=True)

    if mass_as_model:
        clumps = result.instance.clumps.as_model((MassProfile,))

        for clump_index in range(len(result.instance.clumps)):
            if hasattr(result.instance.clumps[clump_index], "mass"):
                clumps[clump_index].mass.centre = result.instance.clumps[
                    clump_index
                ].mass.centre
                clumps[clump_index].mass.einstein_radius = result.model.clumps[
                    clump_index
                ].mass.einstein_radius
                if free_centre:
                    clumps[clump_index].mass.centre = result.model.clumps[
                        clump_index
                    ].mass.centre

    elif light_as_model:
        clumps = result.instance.clumps.as_model((LightProfile,))

        for clump_index in range(len(result.instance.clumps)):
            if clumps[clump_index].light is not None:
                clumps[clump_index].light.centre = result.instance.clumps[
                    clump_index
                ].light.centre
                if free_centre:
                    clumps[clump_index].light.centre = result.model.clumps[
                        clump_index
                    ].light.centre

    else:
        clumps = result.instance.clumps.as_model(())

    return clumps


def lp_chain_tracer_from(light_result, settings_search):
    """
    Output the chain tracer, which is a tracer used for passing the `intensity` values of all linear light profiles
    from the LIGHT PIPELINE to the LIGHT DARK MASS PIPELINE.

    When linear light profiles are used in the LIGHT PIPELINE, their intensities are solved for via linear algebra
    when setting up their corresponding light and mass profiles for the MASS LIGHT DARK PIPELINE.

    This calculation is not numerically accuracy to a small amount (of order 1e-8) in the `intensity` values that
    are solved for. This lack of accuracy will not impact the lens modeling in a noticeable way.

    However, when a pipeline is rerun the `intensity` values that are solved for may change by a small amount,
    changing the unique identifier of the fit, meaning the run does not resume correctly.

    This function outputs the chaining tracer used to pass the `intensity` values to a .json file, or loads
    it from this file if it is already there. This ensures that when a pipeline is rerun, the same `intensity`
    values are always used.
    """

    if settings_search.unique_tag is not None:
        unique_tag = settings_search.unique_tag
    else:
        unique_tag = ""

    lp_chain_tracer_path = path.join(
        conf.instance.output_path,
        settings_search.path_prefix,
        unique_tag,
        "light_dark_lp_chain_tracer.json",
    )

    try:
        lp_chain_tracer = from_json(lp_chain_tracer_path)

    except FileNotFoundError:
        lp_chain_tracer = (
            light_result.max_log_likelihood_fit.model_obj_linear_light_profiles_to_light_profiles
        )

        output_to_json(obj=lp_chain_tracer, file_path=lp_chain_tracer_path)

    return lp_chain_tracer


def basis_no_linear_from(name: str, lp_chain_tracer) -> af.Model:
    """
    Returns a basis containing standard light profiles from a basis (e.g. an MGE) in the LIGHT PIPELINE result,
    for the TOTAL MASS PIPELINE.

    For example, if the light pipeline fits a basis of linear Gaussians, this function will return a basis of standard
    Gaussians where each Gaussian has been converted to standard light profile.

    This is used to fix the lens light subtraction in the TOTAL MASS PIPELINE, as opposed to continue using
    linear light profiles which solve for the intensities for each iteration of the pipeline.

    These profiles will have been converted from standard light profiles / linear light profiles to light and mass
    profiles, where their light profile parameters (e.g. their `centre`, `ell_comps`) are used to set up the parameters
    of the light and mass profile.

    If a linear light profile is passed in, the function will return a light and mass profile where its `intensity` is
    fixed to the solved for value of the maximum log likelihood linear fit.

    The light and mass profiles can also be switched to variants which have a radial gradient in their mass-to-light
    conversion, by setting the `include_mass_to_light_gradient` parameter to `True`.


    Parameters
    ----------
    light_result
        The result of the light pipeline, which determines the light and mass profiles used in the LIGHT DARK
        MASS PIPELINE.
    name
        The name of the light profile in the light pipeline's galaxy model that the model is being created for
        (e.g. `bulge`).

    Returns
    -------
    The light and mass profile for a basis (e.g. an MGE) whose priors are initialized from a previous result.
    """

    try:
        lp_instance = getattr(
            lp_chain_tracer.galaxies[0],
            name,
        )
    except AttributeError:
        return None

    if lp_instance is None:
        return None

    profile_list = lp_instance.profile_list

    lp_model_list = []

    for i, light_profile in enumerate(profile_list):
        lp_model = af.Model(lp.Gaussian)

        lp_model.centre = light_profile.centre
        lp_model.ell_comps = light_profile.ell_comps
        lp_model.intensity = light_profile.intensity
        lp_model.sigma = light_profile.sigma

        lp_model_list += [lp_model]

    return af.Model(Basis, profile_list=lp_model_list)


def mass_light_dark_lmp_from(
    light_result: Result,
    name: str,
    lp_chain_tracer,
    linear_lp_to_standard: bool = False,
    light_is_model: bool = False,
    use_gradient: bool = False,
):
    """
    Returns a light and mass profile from a standard light profile (e.g. a Sersic) in the LIGHT PIPELINE result, for
    the LIGHT DARK MASS PIPELINE.

    For example, if the light pipeline fits a Sersic profile, this function will return a Sersic profile for the LIGHT
    DARK MASS PIPELINE.

    For the light profile, it will by default use a linear light and mass profile, which therefore continues to solve
    for the intensity of the linear light profile via linear algebra. This means that during the fit the `intensity`
    used to compute deflection angles is not the same as the `intensity` parameter solved for in the linear light
    profile. If the input `linear_lp_to_standard` is `True`, the function will instead use a standard light profile.

    The light and mass profiles can also be switched to variants which have a radial gradient in their mass-to-light
    conversion, by setting the `include_mass_to_light_gradient` parameter to `True`.

    Parameters
    ----------
    light_result
        The result of the light pipeline, which determines the light and mass profiles used in the LIGHT DARK
        MASS PIPELINE.
    name
        The name of the light profile in the light pipeline's galaxy model that the model is being created for
    light_is_model
        If `True`, the light profile is passed as a model component, else it is a fixed instance.

    Returns
    -------
    The light and mass profile for a standard light profile whose priors are initialized from a previous result.
    """

    lp_instance = getattr(light_result.instance.galaxies.lens, name)
    lp_model = getattr(light_result.model.galaxies.lens, name)

    try:
        is_linear = False

        if not use_gradient:
            lp_to_lmp_dict = {lp.Sersic: lmp.Sersic}
        else:
            lp_to_lmp_dict = {lp.Sersic: lmp.SersicGradient}

        lmp_model = lp_to_lmp_dict[type(lp_instance)]

    except KeyError:
        if not linear_lp_to_standard:
            if not use_gradient:
                lp_linear_to_lmp_dict = {lp_linear.Sersic: lmp_linear.Sersic}
            else:
                lp_linear_to_lmp_dict = {lp_linear.Sersic: lmp_linear.SersicGradient}
        else:
            if not use_gradient:
                lp_linear_to_lmp_dict = {lp_linear.Sersic: lmp.Sersic}
            else:
                lp_linear_to_lmp_dict = {lp_linear.Sersic: lmp.SersicGradient}

        is_linear = True
        lmp_model = lp_linear_to_lmp_dict[type(lp_instance)]

    lmp_model = af.Model(lmp_model)

    if light_is_model:
        lmp_model.take_attributes(source=lp_model)
    else:
        lmp_model.take_attributes(source=lp_instance)

    if is_linear:
        fit = light_result.max_log_likelihood_fit
        lp_solved = getattr(lp_chain_tracer.galaxies[0], name)

        lmp_model.intensity = lp_solved.intensity

    return lmp_model


def mass_light_dark_basis_from(
    light_result: Result,
    name: str,
    lp_chain_tracer,
    linear_lp_to_standard: bool = False,
    use_gradient: bool = False,
) -> af.Model:
    """
    Returns a basis containing light and mass profiles from a basis (e.g. an MGE) in the LIGHT PIPELINE result, for the
    LIGHT DARK MASS PIPELINE.

    For example, if the light pipeline fits a basis of Gaussians, this function will return a basis of Gaussians for the
    LIGHT DARK MASS PIPELINE where each Gaussian has been converted to a light and mass profile.

    These profiles will have been converted from standard light profiles / linear light profiles to light and mass
    profiles, where their light profile parameters (e.g. their `centre`, `ell_comps`) are used to set up the parameters
    of the light and mass profile.

    For the light profile, it will by default use a linear light and mass profile, which therefore continues to solve
    for the intensity of the linear light profile via linear algebra. This means that during the fit the `intensity`
    used to compute deflection angles is not the same as the `intensity` parameter solved for in the linear light
    profile. If the input `linear_lp_to_standard` is `True`, the function will instead use a standard light profile.

    The light and mass profiles can also be switched to variants which have a radial gradient in their mass-to-light
    conversion, by setting the `include_mass_to_light_gradient` parameter to `True`.

    Parameters
    ----------
    light_result
        The result of the light pipeline, which determines the light and mass profiles used in the LIGHT DARK
        MASS PIPELINE.
    name
        The name of the light profile in the light pipeline's galaxy model that the model is being created for
        (e.g. `bulge`).

    Returns
    -------
    The light and mass profile for a basis (e.g. an MGE) whose priors are initialized from a previous result.
    """

    try:
        lp_instance = getattr(
            lp_chain_tracer.galaxies[0],
            name,
        )
    except AttributeError:
        return None

    profile_list = lp_instance.profile_list

    lmp_model_list = []

    for i, light_profile in enumerate(profile_list):
        if not linear_lp_to_standard:
            if not use_gradient:
                lmp_model = af.Model(lmp_linear.Gaussian)
            else:
                lmp_model = af.Model(lmp_linear.GaussianGradient)
        else:
            if not use_gradient:
                lmp_model = af.Model(lmp.Gaussian)
            else:
                lmp_model = af.Model(lmp.GaussianGradient)

        lmp_model.centre = light_profile.centre
        lmp_model.ell_comps = light_profile.ell_comps
        lmp_model.intensity = light_profile.intensity
        lmp_model.sigma = light_profile.sigma

        lmp_model_list += [lmp_model]

        if not use_gradient:
            lmp_model.mass_to_light_ratio = lmp_model_list[0].mass_to_light_ratio
        else:
            lmp_model.mass_to_light_ratio_base = lmp_model_list[
                0
            ].mass_to_light_ratio_base
            lmp_model.mass_to_light_gradient = lmp_model_list[0].mass_to_light_gradient

    return af.Model(Basis, profile_list=lmp_model_list)


def mass_light_dark_from(
    light_result: Result,
    name: str,
    lp_chain_tracer,
    linear_lp_to_standard: bool = False,
    light_is_model: bool = False,
    use_gradient: bool = False,
) -> Optional[af.Model]:
    """
    Returns light and mass profiles from the LIGHT PIPELINE result, for the LIGHT DARK MASS PIPELINE.

    For example, if the light pipeline fits Sersic profiles, this function will return Sersic profiles for the LIGHT
    DARK MASS PIPELINE.

    These profiles will have been converted from standard light profiles / linear light profiles to light and mass
    profiles, where their light profile parameters (e.g. their `centre`, `ell_comps`) are used to set up the parameters
    of the light and mass profile.

    If a linear light profile is passed in, the function will return a mass profile where its `intensity` is
    fixed to the solved for value of the maximum log likelihood linear fit.

    For the light profile, it will by default use a linear light and mass profile, which therefore continues to solve
    for the intensity of the linear light profile via linear algebra. This means that during the fit the `intensity`
    used to compute deflection angles is not the same as the `intensity` parameter solved for in the linear light
    profile. If the input `linear_lp_to_standard` is `True`, the function will instead use a standard light profile.

    This function supports the input of a basis (e.g. an MGE), converting every individual standard light or
    linear light profile in the basis to a light and mass profile.

    The light and mass profiles can also be switched to variants which have a radial gradient in their mass-to-light
    conversion, by setting the `include_mass_to_light_gradient` parameter to `True`.


    Parameters
    ----------
    light_result
        The result of the light pipeline, which determines the light and mass profiles used in the LIGHT DARK
        MASS PIPELINE.
    name
        The name of the light profile in the light pipeline's galaxy model that the model is being created for
        (e.g. `bulge`).
    linear_lp_to_standard
        If `True`, the light profile is passed as a standard component, else it is a linear component. The
        mass always uses the `interity` parameter of the linear light profile solved for in the previous pipeline.
    light_is_model
        If `True`, the light profile is passed as a model component, else it is a fixed instance. For a basis
        (e.g. an MGE) this feature is not used due to the large number of profiles in the basis.
    use_gradient
        If `True`, the light and mass profiles is switched to a variant which includes additional parameters that
        model a radial gradient in the mass-to-light ratio.

    Returns
    -------
    The light and mass profile whose priors are initialized from a previous result.
    """

    lp_instance = getattr(light_result.instance.galaxies.lens, name)

    if lp_instance is None:
        return None

    if not isinstance(lp_instance, Basis):
        return mass_light_dark_lmp_from(
            light_result=light_result,
            name=name,
            lp_chain_tracer=lp_chain_tracer,
            linear_lp_to_standard=linear_lp_to_standard,
            light_is_model=light_is_model,
            use_gradient=use_gradient,
        )
    return mass_light_dark_basis_from(
        light_result=light_result,
        lp_chain_tracer=lp_chain_tracer,
        name=name,
        use_gradient=use_gradient,
    )


def link_ratios(link_mass_to_light_ratios: bool, light_result, bulge, disk):
    """
    Links the mass to light ratios and gradients of the bulge and disk profiles in the MASS LIGHT DARK PIPELINE.

    The following use cases are supported:

    1) Link the mass to light ratios of the bulge and diskwhen they are both sersic light and mass profile.
    2) Link the mass to light ratios of the bulge and disk when for a basis (e.g. an MGE).
    3) Does approrpiate linking with the gradient of the mass-to-light ratio for the basis profiles.

    Parameters
    ----------
    link_mass_to_light_ratios
        Whether the mass-to-light ratios of the bulge and disk profiles are linked.
    light_result
        The result of the light pipeline, which determines the light and mass profiles used in the MASS LIGHT DARK
        PIPELINE.
    bulge
        The bulge model light profile.
    disk
        The disk model light profile.

    Returns
    -------
    The bulge and disk profiles with the mass-to-light ratios linked.
    """

    if bulge is None or disk is None:
        return bulge, disk

    if not link_mass_to_light_ratios:
        return bulge, disk

    bulge_instance = getattr(light_result.instance.galaxies.lens, "bulge")

    if not isinstance(bulge_instance, Basis):
        bulge.mass_to_light_ratio = disk.mass_to_light_ratio

        return bulge, disk

    for bulge_lp in bulge.profile_list:
        try:
            bulge_lp.mass_to_light_ratio = disk.profile_list[0].mass_to_light_ratio
        except AttributeError:
            bulge_lp.mass_to_light_ratio_base = disk.profile_list[
                0
            ].mass_to_light_ratio_base
            bulge_lp.mass_to_light_gradient = disk.profile_list[
                0
            ].mass_to_light_gradient

    return bulge, disk



import numpy as np
from scipy.special import factorial, genlaguerre
from typing import Optional, Tuple

import autoarray as aa


from autogalaxy.profiles.light.decorators import (
    check_operated_only,
)
from autogalaxy.profiles.light.standard.shapelets.abstract import AbstractShapelet


class ShapeletPolar(AbstractShapelet):
    def __init__(
        self,
        n: int,
        m: int,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 1.0,
        beta: float = 1.0,
    ):
        """
        Shapelets where the basis function is defined according to a Polar (r,theta) grid of coordinates.

        Shapelets are defined according to:

          https://arxiv.org/abs/astro-ph/0105178

        Shapelets are described in the context of strong lens modeling in:

          https://ui.adsabs.harvard.edu/abs/2016MNRAS.457.3066T/abstract

        Parameters
        ----------
        n
            The n order of the shapelets basis function.
        m
            The m order of the shapelets basis function in the x-direction.
        centre
            The (y,x) arc-second coordinates of the profile (shapelet) centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        beta
            The characteristic length scale of the shapelet basis function, defined in arc-seconds.
        """

        self.n = n
        self.m = m

        super().__init__(
            centre=centre, ell_comps=ell_comps, beta=beta, intensity=intensity
        )

    @property
    def coefficient_tag(self) -> str:
        return f"n_{self.n}_m_{self.m}"

    @aa.over_sample
    @aa.grid_dec.to_array
    @check_operated_only
    @aa.grid_dec.transform
    @aa.grid_dec.relocate_to_radial_minimum
    def image_2d_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None, **kwargs
    ) -> np.ndarray:
        """
        Returns the Polar Shapelet light profile's 2D image from a 2D grid of Polar (y,x) coordinates.

        If the coordinates have not been transformed to the profile's geometry (e.g. translated to the
        profile `centre`), this is performed automatically.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        image
            The image of the Polar Shapelet evaluated at every (y,x) coordinate on the transformed grid.
        """

        laguerre = genlaguerre(n=(self.n - np.abs(self.m)) / 2.0, alpha=np.abs(self.m))

        const = (
            ((-1) ** ((self.n - np.abs(self.m)) // 2))
            * np.sqrt(
                factorial((self.n - np.abs(self.m)) // 2)
                / factorial((self.n + np.abs(self.m)) // 2)
            )
            / self.beta
            / np.sqrt(np.pi)
        )

        rsq = (grid[:, 0] ** 2 + grid[:, 1] ** 2) / self.beta**2
        theta = np.arctan2(grid[:, 1], grid[:, 0])
        radial = rsq ** (abs(self.m / 2.0)) * np.exp(-rsq / 2.0) * laguerre(rsq)

        if self.m == 0:
            azimuthal = 1
        elif self.m > 0:
            azimuthal = np.sin((-1) * self.m * theta)
        else:
            azimuthal = np.cos((-1) * self.m * theta)

        return const * radial * azimuthal


class ShapeletPolarSph(ShapeletPolar):
    def __init__(
        self,
        n: int,
        m: int,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 1.0,
        beta: float = 1.0,
    ):
        """
        Shapelets where the basis function is defined according to a Polar (r,theta) grid of coordinates.

        Shapelets are defined according to:

          https://arxiv.org/abs/astro-ph/0105178

        Shapelets are described in the context of strong lens modeling in:

          https://ui.adsabs.harvard.edu/abs/2016MNRAS.457.3066T/abstract

        Parameters
        ----------
        n_y
            The order of the shapelets basis function in the y-direction.
        n_x
            The order of the shapelets basis function in the x-direction.
        centre
            The (y,x) arc-second coordinates of the profile (shapelet) centre.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        beta
            The characteristic length scale of the shapelet basis function, defined in arc-seconds.
        """

        super().__init__(
            n=n,
            m=m,
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            beta=beta,
        )



import numpy as np
from scipy.integrate import quad
from typing import Optional, Tuple

import autoarray as aa

from autogalaxy.operate.image import OperateImage
from autogalaxy.profiles.geometry_profiles import EllProfile


class LightProfile(EllProfile, OperateImage):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
    ):
        """
        Abstract base class for an elliptical light-profile.

        Each light profile has an analytic equation associated with it that describes its 1D surface brightness.

        Given an input grid of 1D or 2D (y,x) coordinates the light profile can be used to evaluate its surface
        brightness in 1D or as a 2D image.

        Associated with a light profile is a spherical or elliptical geometry, which describes its `centre` of
        emission and ellipticity. Geometric transformations are performed by decorators linked to the **PyAutoArray**
        `geometry` package.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system (see the module
            `autogalaxy -> convert.py` for the convention).
        """
        super().__init__(centre=centre, ell_comps=ell_comps)
        self.intensity = intensity

    @property
    def coefficient_tag(self) -> str:
        return ""

    def image_2d_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None, **kwargs
    ) -> aa.Array2D:
        """
        Returns the light profile's 2D image from a 2D grid of Cartesian (y,x) coordinates, which may have been
        transformed using the light profile's geometry.

        If the coordinates have not been transformed to the profile's geometry (e.g. translated to the
        profile `centre`), this is performed automatically.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        image
            The image of the `LightProfile` evaluated at every (y,x) coordinate on the transformed grid.
        """
        raise NotImplementedError()

    def image_2d_via_radii_from(self, grid_radii: np.ndarray) -> np.ndarray:
        """
        Returns the light profile's 2D image from a 1D grid of coordinates which are the radial distance of each
        coordinate from the light profile `centre`.

        Parameters
        ----------
        grid_radii
            The radial distances from the centre of the profile, for each coordinate on the grid.
        """
        raise NotImplementedError()

    @aa.grid_dec.project_grid
    def image_1d_from(
        self, grid: aa.type.Grid1D2DLike, **kwargs
    ) -> aa.type.Grid1D2DLike:
        """
        Returns the light profile's 1D image from a grid of Cartesian coordinates, which may have been
        transformed using the light profile's geometry.

        If a 1D grid is input the image is evaluated every coordinate on the grid. If a 2D grid is input, this is
        converted to a 1D grid by aligning with the major-axis of the light profile's elliptical geometry.

        Internally, this function uses a 2D grid to compute the image, which is mapped to a 1D data structure on return
        via the `project_grid` decorator. This avoids code repetition by ensuring that light profiles only use
        their `image_2d_from()`  function to evaluate their image.

        Parameters
        ----------
        grid
            A 1D or 2D grid of coordinates which are used to evaluate the light profile in 1D.

        Returns
        -------
        image
            The 1D image of the light profile evaluated at every (x,) coordinate on the 1D transformed grid.
        """
        return self.image_2d_from(grid=grid, **kwargs)

    def luminosity_within_circle_from(self, radius: float) -> float:
        """
        Integrate the light profile to compute the total luminosity within a circle of specified radius. This is
        centred on the light profile's `centre`.

        The `intensity` of a light profile is in dimension units, which are given physical meaning when the light
        profile is compared to data with physical units. The luminosity output by this function therefore is also
        dimensionless until compared to data.

        Parameters
        ----------
        radius
            The radius of the circle to compute the dimensionless luminosity within.
        """

        return quad(func=self.luminosity_integral, a=0.0, b=radius)[0]

    def luminosity_integral(self, x: np.ndarray) -> np.ndarray:
        """
        Routine to integrate the luminosity of an elliptical light profile.

        The axis ratio is set to 1.0 for computing the luminosity within a circle

        Parameters
        ----------
        x
            The 1D (x) radial coordinates where the luminosity integral is evaluated.
        """
        return 2 * np.pi * x * self.image_2d_via_radii_from(x)

    @property
    def half_light_radius(self) -> float:
        if hasattr(self, "effective_radius"):
            return self.effective_radius

    @property
    def _intensity(self):
        return self.intensity



from autoarray.geometry import geometry_util as geometry
from autoarray.mask import mask_1d_util as mask_1d
from autoarray.mask import mask_2d_util as mask_2d
from autoarray.structures.arrays import array_1d_util as array_1d
from autoarray.structures.arrays import array_2d_util as array_2d
from autoarray.structures.grids import grid_1d_util as grid_1d
from autoarray.structures.grids import grid_2d_util as grid_2d
from autoarray.structures.grids import sparse_2d_util as sparse
from autoarray.fit import fit_util as fit
from autoarray.inversion.pixelization.mesh import mesh_util as mesh
from autoarray.inversion.pixelization.mappers import mapper_util as mapper
from autoarray.inversion.regularization import regularization_util as regularization
from autoarray.inversion.inversion import inversion_util as inversion
from autoarray.inversion.inversion.imaging import (
    inversion_imaging_util as inversion_imaging,
)
from autoarray.inversion.inversion.interferometer import (
    inversion_interferometer_util as inversion_interferometer,
)
from autoarray.operators import transformer_util as transformer
from autoarray.util import misc_util as misc

from autogalaxy.util import error_util as error
from autogalaxy.analysis import chaining_util as chaining
